#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess
import shlex

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk, Gio, Gtk, GLib  # type: ignore


logger = logging.getLogger('menulibre')


class FileHandler:

    def open_folder(self, path):
        if not os.path.isdir(path):
            path = os.path.dirname(path)

        uri = "file://%s" % path

        if "show_uri_on_window" in dir(Gtk):
            Gtk.show_uri_on_window(None, uri, 0)
        else:
            Gtk.show_uri(None, uri, 0)

        return True

    def open_editor(self, path):
        if os.path.isdir(path):
            return self.open_folder(path)

        file = self._get_file(path)

        if not self._file_is_writable(file):
            admin = self._get_preferred_admin_editor()
            if admin is not None:
                return admin.launch(
                    [self._get_file("admin://%s" % file.get_path())])

            logging.warning(
                "Could not find a text editor supporting admin://")

            root = self._get_root_editor()
            if root is not None:
                return root.launch([file])

            logging.warning(
                "Could not find a text editor supporting pkexec")

        editor = self._get_editor()
        if editor is not None:
            return editor.launch([file])

        logging.warning(
            "Could not find a supported text editor")

        return False

    def copy_to_clipboard(self, path):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(path, -1)

    def _get_file(self, path):
        if "://" in path:
            return Gio.File.new_for_uri(path)
        else:
            return Gio.File.new_for_path(path)

    def _file_is_writable(self, file: Gio.File):
        try:
            info = file.query_info(
                Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE,
                Gio.FileQueryInfoFlags.NONE,
                None)
            return info.get_attribute_boolean(
                Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE)
        except BaseException:
            return False

    def _get_preferred_admin_editor(self):
        default = self._get_editor()
        if self._editor_supports_admin_protocol(default):
            return default
        for editor in self._get_editors():
            if self._editor_supports_admin_protocol(editor):
                return editor
        return None

    def _get_editor(self):
        info = Gio.AppInfo.get_default_for_type("text/plain", False)
        if info is not None:
            return info
        return None

    def _get_editors(self):
        execs = {}
        infos = Gio.AppInfo.get_all_for_type("text/plain")
        for info in infos:
            appid = info.get_id()
            if appid is not None:
                appid = appid[:-8]
                execs[appid] = info
        return execs

    def _editor_supports_admin_protocol(self, editor):
        return editor in [
            "gedit",
            "pluma"
        ]

    def _get_keyfile(self, editor: Gio.AppInfo):
        app_id = editor.get_id()
        fname = app_id

        keyfile = GLib.KeyFile.new()

        data_dirs = [GLib.get_user_data_dir()] + GLib.get_system_data_dirs()
        for data_dir in data_dirs:
            for path, directories, files in os.walk(data_dir):
                if fname in files:
                    filename = os.path.join(data_dir, path, fname)
                    if keyfile.load_from_file(
                            filename, GLib.KeyFileFlags.NONE):
                        return keyfile

        return None

    def _get_pkexec_editor(self, editor: Gio.AppInfo, pkexec: str):
        keyfile = self._get_keyfile(editor)
        if keyfile is None:
            return None

        try:
            commandline = keyfile.get_string("Desktop Entry", "Exec")
            command = shlex.join([pkexec] + shlex.split(commandline))
            keyfile.set_string("Desktop Entry", "Exec", command)
        except GLib.Error:
            return None

        desktop = Gio.DesktopAppInfo.new_from_keyfile(keyfile)
        if desktop is None:
            return None

        return desktop

    def _get_root_editor(self):
        pkexec_bin = GLib.find_program_in_path("pkexec")
        if pkexec_bin is None:
            logging.warning(
                "Could not find pkexec to for executing root editors")
            return None
        if self._get_display_server() == "wayland":
            logging.warning(
                "pkexec is not supported under Wayland")
            return None
        pkexecs = self._get_pkexecs()
        default = self._get_editor()
        preferred = None
        for appid, editor in self._get_editors().items():
            for pkexec in pkexecs:
                if pkexec == appid or pkexec == "org.freedesktop.policykit.pkexec.%s" % appid:
                    if default is not None and editor.get_id() == default.get_id():
                        return self._get_pkexec_editor(editor, pkexec_bin)
                    if preferred is None:
                        preferred = editor
        if preferred is not None:
            return self._get_pkexec_editor(preferred, pkexec_bin)
        if default is not None and default.get_id() not in ["code.desktop"]:
            return self._get_pkexec_editor(default, pkexec_bin)
        logging.warning(
            "Could not find a text editor supporting pkexec")
        return None

    def _get_display_server(self):
        wayland = os.getenv("WAYLAND_DISPLAY")
        if wayland is not None:
            return "wayland"
        x11 = os.getenv("DISPLAY")
        if x11 is not None:
            return "x11"
        logging.warning(
            "Could not determine display server. Assuming x11")
        return "x11"

    def _get_pkexecs(self):
        try:
            output = subprocess.check_output(
                ["pkaction"], stderr=subprocess.STDOUT)
        except Exception as e:
            output = e.output  # type: ignore
        try:
            return output.decode('utf-8').split("\n")
        except BaseException:
            return []



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from locale import gettext as _

from gi.repository import Gtk  # type: ignore

import menulibre_lib

import logging
logger = logging.getLogger('menulibre')


class AboutDialog(Gtk.AboutDialog):
    def __init__(self, parent, use_headerbar):
        Gtk.AboutDialog.__init__(self, use_header_bar=use_headerbar)
        authors = ["Sean Davis"]
        documenters = ["Sean Davis"]

        # Translators: About Dialog, window title.
        self.set_title(_("About MenuLibre"))
        self.set_program_name("MenuLibre")
        self.set_logo_icon_name("menulibre")
        self.set_copyright("Copyright © 2012-2024 Sean Davis")
        self.set_authors(authors)
        self.set_documenters(documenters)
        self.set_website("https://github.com/bluesabre/menulibre")
        self.set_version(menulibre_lib.get_version())

        # Connect the signal to destroy the AboutDialog when Close is clicked.
        self.connect("response", self.about_close_cb)
        self.set_transient_for(parent)

        # Fix weird bug with duplicate buttons
        if use_headerbar:
            headerbar = self.get_header_bar()
            for child in headerbar.get_children():
                if isinstance(child, Gtk.Button):
                    child.destroy()

    def about_close_cb(self, widget, response):
        """Destroy the AboutDialog when it is closed."""
        widget.destroy()


class HelpDialog(Gtk.MessageDialog):
    def __init__(self, parent, use_headerbar):
        # Translators: Help Dialog, window title.
        title = _("Online Documentation")
        # Translators: Help Dialog, primary text.
        primary = _("Do you want to read the MenuLibre manual online?")
        # Translators: Help Dialog, secondary text.
        secondary = _("You will be redirected to the documentation website "
                      "where the help pages are maintained.")
        buttons = [
            # Translators: Help Dialog, cancel button.
            (_("Cancel"), Gtk.ResponseType.CANCEL),
            # Translators: Help Dialog, confirmation button. Navigates to
            # online documentation.
            (_("Read Online"), Gtk.ResponseType.OK)
        ]

        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.QUESTION,
                                   buttons=Gtk.ButtonsType.NONE,
                                   text=primary, use_header_bar=False)
        self.set_title(title)
        self.format_secondary_markup(secondary)
        for button in buttons:
            self.add_button(button[0], button[1])

        self.connect("response", self.response_cb)

    def response_cb(self, widget, response):
        if response == Gtk.ResponseType.OK:
            url = "https://github.com/bluesabre/menulibre/wiki"
            logger.debug("Navigating to help page, %s" % url)
            menulibre_lib.show_uri(self.get_transient_for(), url)
        widget.destroy()


class SaveOnCloseDialog(Gtk.MessageDialog):
    def __init__(self, parent, use_headerbar):
        # Translators: Save On Close Dialog, window title.
        title = _("Save Changes")
        # Translators: Save On Close Dialog, primary text.
        primary = _("Do you want to save the changes before closing?")
        # Translators: Save On Close Dialog, secondary text.
        secondary = _("If you don't save the launcher, all the changes "
                      "will be lost.")
        buttons = [
            # Translators: Save On Close Dialog, don't save, then close.
            (_("Don't Save"), Gtk.ResponseType.NO),
            # Translators: Save On Close Dialog, don't save, cancel close.
            (_("Cancel"), Gtk.ResponseType.CANCEL),
            # Translators: Save On Close Dialog, do save, then close.
            (_("Save"), Gtk.ResponseType.YES)
        ]

        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.QUESTION,
                                   buttons=Gtk.ButtonsType.NONE,
                                   text=primary, use_header_bar=False)
        self.set_title(title)
        self.format_secondary_markup(secondary)
        for button in buttons:
            self.add_button(button[0], button[1])


class SaveOnLeaveDialog(Gtk.MessageDialog):
    def __init__(self, parent, use_headerbar):
        # Translators: Save On Leave Dialog, window title.
        title = _("Save Changes")
        # Translators: Save On Leave Dialog, primary text.
        primary = _("Do you want to save the changes before leaving this "
                    "launcher?")
        # Translators: Save On Leave Dialog, primary text.
        secondary = _("If you don't save the launcher, all the changes "
                      "will be lost.")
        buttons = [
            # Translators: Save On Leave Dialog, don't save, then leave.
            (_("Don't Save"), Gtk.ResponseType.NO),
            # Translators: Save On Leave Dialog, don't save, cancel leave.
            (_("Cancel"), Gtk.ResponseType.CANCEL),
            # Translators: Save On Leave Dialog, do save, then leave.
            (_("Save"), Gtk.ResponseType.YES)
        ]

        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.QUESTION,
                                   buttons=Gtk.ButtonsType.NONE,
                                   text=primary, use_header_bar=False)
        self.set_title(title)
        self.format_secondary_markup(secondary)
        for button in buttons:
            self.add_button(button[0], button[1])


class DeleteDialog(Gtk.MessageDialog):
    def __init__(self, parent, primary, use_headerbar):
        # Translations: Delete Dialog, secondary text. Notifies user that
        # the file cannot be restored once deleted.
        secondary = _("This cannot be undone.")
        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.QUESTION,
                                   buttons=Gtk.ButtonsType.OK_CANCEL,
                                   text=primary, use_header_bar=False)
        self.format_secondary_markup(secondary)


class RevertDialog(Gtk.MessageDialog):
    def __init__(self, parent, use_headerbar):
        # Translators: Revert Dialog, window title.
        title = _("Restore Launcher")
        # Translators: Revert Dialog, primary text. Confirmation to revert
        # all changes since the last file save.
        primary = _("Are you sure you want to restore this launcher?")
        # Translators: Revert Dialog, secondary text.
        secondary = _("All changes since the last saved state will be lost "
                      "and cannot be restored automatically.")
        buttons = [
            # Translators: Revert Dialog, cancel button.
            (_("Cancel"), Gtk.ResponseType.CANCEL),
            # Translators: Revert Dialog, confirmation button.
            (_("Restore Launcher"), Gtk.ResponseType.OK)
        ]

        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.QUESTION,
                                   buttons=Gtk.ButtonsType.NONE,
                                   text=primary, use_header_bar=False)
        self.set_title(title)
        self.format_secondary_markup(secondary)
        for button in buttons:
            self.add_button(button[0], button[1])


class LauncherRemovedDialog(Gtk.MessageDialog):
    def __init__(self, parent, use_headerbar):
        # Translators: Launcher Removed Dialog, primary text. Indicates that
        # the selected application is no longer installed.
        primary = _("No Longer Installed")
        # Translators: Launcher Removed Dialog, secondary text.
        secondary = _("This launcher has been removed from the "
                      "system.\nSelecting the next available item.")

        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.INFO,
                                   buttons=Gtk.ButtonsType.OK,
                                   text=primary, use_header_bar=False)
        self.format_secondary_markup(secondary)


class NotFoundInPathDialog(Gtk.MessageDialog):
    def __init__(self, parent, command, use_headerbar):
        # Translators: Not Found In PATH Dialog, primary text. Indicates
        # that the provided script was not found in any PATH directory.
        primary = _("Could not find \"%s\" in your PATH.") % command

        path = os.getenv("PATH", "").split(":")
        secondary = "<b>PATH:</b>\n%s" % "\n".join(path)
        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.ERROR,
                                   buttons=Gtk.ButtonsType.OK,
                                   text=primary, use_header_bar=False)
        self.format_secondary_markup(secondary)
        self.connect("response", self.response_cb)

    def response_cb(self, widget, user_data):
        widget.destroy()


class SaveErrorDialog(Gtk.MessageDialog):
    def __init__(self, parent, filename, use_headerbar):
        # Translators: Save Error Dialog, primary text.
        primary = _("Failed to save \"%s\".") % filename
        # Translators: Save Error Dialog, secondary text.
        secondary = \
            _("Do you have write permission to the file and directory?")

        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.ERROR,
                                   buttons=Gtk.ButtonsType.OK,
                                   text=primary, use_header_bar=False)
        self.format_secondary_markup(secondary)
        self.connect("response", self.response_cb)

    def response_cb(self, widget, user_data):
        widget.destroy()



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from locale import gettext as _
from .LabelWithHidingButton import LabelWithHidingButton

import menulibre_lib

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango, GObject  # type: ignore


class FieldLabel(LabelWithHidingButton):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self, label, key_name, description,
                 help_text=None, help_url=None):
        super().__init__(
            label=label,
            icon_name="dialog-question-symbolic",
            icon_size=Gtk.IconSize.BUTTON)
        self._value = None

        label = self.get_label()
        label.set_ellipsize(Pango.EllipsizeMode.NONE)

        button = self.get_button()
        button.set_tooltip_markup(
            _("More information about <i>%s</i>") %
            key_name)
        button.connect(
            "clicked",
            self._button_clicked_cb,
            key_name,
            description,
            help_text,
            help_url)

    def _button_clicked_cb(self, widget, key_name,
                           description, help_text, help_url):
        dlg = FieldInfo(
            self.get_toplevel(),
            key_name=key_name,
            description=description,
            help_text=help_text,
            help_url=help_url)
        dlg.show()


class FieldInfo(Gtk.MessageDialog):
    def __init__(self, parent, key_name, description, help_text, help_url):
        primary = key_name
        secondary = description

        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.ERROR,
                                   buttons=Gtk.ButtonsType.CLOSE,
                                   text=primary, use_header_bar=False)
        self.format_secondary_markup(secondary)

        message_area = self.get_content_area().get_children()[0]
        if isinstance(message_area.get_children()[0], Gtk.Image):  # type: ignore
            message_area.get_children()[0].destroy()  # type: ignore

        if help_text:
            button = self.add_button(help_text, Gtk.ResponseType.HELP)
            bbox = self.get_action_area()
            bbox.set_child_secondary(button, True)  # type: ignore

        self.connect("response", self.response_cb, help_url)

    def response_cb(self, widget, response, help_url):
        if response == Gtk.ResponseType.HELP:
            menulibre_lib.show_uri(self.get_transient_for(), help_url)
            return
        widget.destroy()



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#   Copyright (C) 2016-2018 OmegaPhil <OmegaPhil@startmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from .util import getCurrentDesktop, getProcessList
from .util import MenuItemTypes, check_keypress, getRelativeName, getRelatedKeys
from .Headerbar import Headerbar
from .Toolbar import Toolbar
from .ApplicationEditor import ApplicationEditor
from .CategoryEditor import category_lookup
from . import MenuEditor
from . import MenulibreXdg, util, ParsingErrorsDialog
from . import MenulibreTreeview, MenulibreHistory, Dialogs
import os
import shlex
import sys

import subprocess
import tempfile

from locale import gettext as _

from gi import require_version
require_version('Gtk', '3.0')
from gi.repository import Gio, GLib, GObject, Gtk, Gdk  # type: ignore


logger = logging.getLogger('menulibre')

session = os.getenv("DESKTOP_SESSION", "")
root = os.getuid() == 0

current_desktop = getCurrentDesktop()


class MenulibreWindow(Gtk.ApplicationWindow):
    """The Menulibre application window."""

    __gsignals__ = {
        'about': (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE,
                  (GObject.TYPE_BOOLEAN,)),
        'help': (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE,
                 (GObject.TYPE_BOOLEAN,)),
        'quit': (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE,
                 (GObject.TYPE_BOOLEAN,)),
        'action-enabled': (GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE,
                           (GObject.TYPE_STRING, GObject.TYPE_BOOLEAN,)),
    }

    def __init__(self, app, headerbar_pref=True):
        """Initialize the Menulibre application."""
        self.root_lockout()

        self.action_items = dict()

        # Set up History
        self.history = MenulibreHistory.History()
        self.history.connect('undo-changed', self.on_undo_changed)
        self.history.connect('redo-changed', self.on_redo_changed)
        self.history.connect('revert-changed', self.on_revert_changed)

        # Steal the window contents for the GtkApplication.
        self.configure_application_window(app)

        self.values = dict()

        # Set up the actions and toolbar
        self.configure_application_actions()

        add_menu = self.get_add_menu()

        self.search_box = Gtk.SearchEntry.new()
        self.search_box.set_placeholder_text(_("Search"))
        self.search_box.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, "edit-find-symbolic")
        self.search_box.connect('icon-press', self.on_search_cleared)

        self.use_headerbar = headerbar_pref
        if headerbar_pref:
            self.configure_headerbar(add_menu)
        else:
            self.configure_application_toolbar(add_menu)

        # Configure events for the headerbar or toolbar (they have the same
        # setup)
        self.connect_toolbar()

        self.configure_css()

        # Set up the application browser
        self.configure_application_treeview()

        # Determining paths of bad desktop files GMenu can't load - if some are
        # detected, alerting user via InfoBar
        self.bad_desktop_files = util.determine_bad_desktop_files()
        if self.bad_desktop_files:
            self.configure_application_bad_desktop_files_infobar()

        self.configure_menu_restart_infobar()

        self.show_all()

        self.on_apps_browser_cursor_changed(None, None)

    def connect_toolbar(self):
        self.insert_action_item('add_button', self.add_button)

        self.save_button.connect(
            "clicked",
            self.activate_action_cb,
            'save_launcher')
        self.save_button.set_sensitive(False)
        self.insert_action_item('save_launcher', self.save_button)

        self.undo_button.connect("clicked", self.activate_action_cb, 'undo')
        self.undo_button.set_sensitive(False)
        self.insert_action_item('undo', self.undo_button)

        self.redo_button.connect("clicked", self.activate_action_cb, 'redo')
        self.redo_button.set_sensitive(False)
        self.insert_action_item('redo', self.redo_button)

        self.revert_button.connect(
            "clicked", self.activate_action_cb, 'revert')
        self.revert_button.set_sensitive(False)
        self.insert_action_item('revert', self.revert_button)

        self.execute_button.connect(
            "clicked", self.activate_action_cb, 'execute')
        self.insert_action_item('execute', self.execute_button)

        self.delete_button.connect(
            "clicked", self.activate_action_cb, 'delete')
        self.delete_button.set_sensitive(False)
        self.insert_action_item('delete', self.delete_button)

    def insert_action_item(self, key, widget):
        if key not in self.action_items.keys():
            self.action_items[key] = []
        self.action_items[key].append(widget)

    def get_add_menu(self):
        menu = Gio.Menu.new()

        menu_items = {
            'app.add_launcher': _("Add _Launcher"),
            'app.add_directory': _("Add _Directory"),
            'app.add_separator': _("Add _Separator")
        }

        for action_name, label in menu_items.items():
            menu.append(label, action_name)

        return menu

    def root_lockout(self):
        if root:
            # Translators: This error is displayed when the application is run
            # as a root user. The application exits once the dialog is
            # dismissed.
            primary = _("MenuLibre cannot be run as root.")

            docs_url = "https://github.com/bluesabre/menulibre/wiki/Frequently-Asked-Questions"

            # Translators: This link goes to the online documentation with more
            # information.
            secondary = _("Please see the "
                          "<a href='%s'>online documentation</a> "
                          "for more information.") % docs_url

            dialog = Gtk.MessageDialog(
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text=primary,
                use_header_bar=False)
            dialog.format_secondary_markup(secondary)
            dialog.run()
            sys.exit(1)

    def menu_load_failure(self):
        primary = _("MenuLibre failed to load.")

        docs_url = "https://github.com/bluesabre/menulibre/wiki/Frequently-Asked-Questions"

        # Translators: This link goes to the online documentation with more
        # information.
        secondary = _("The default menu could not be found. Please see the "
                      "<a href='%s'>online documentation</a> "
                      "for more information.") % docs_url

        secondary += "\n\n<big><b>%s</b></big>" % _("Diagnostics")

        diagnostics = util.getMenuDiagnostics()
        for k, v in diagnostics.items():
            secondary += "\n<b>%s</b>: %s" % (k, v)

        dialog = Gtk.MessageDialog(
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.CLOSE,
            text=primary,
            use_header_bar=False)
        dialog.format_secondary_markup(secondary)

        label = self.find_secondary_label(dialog)
        if label is not None:
            label.set_selectable(True)

        dialog.run()
        sys.exit(1)

    def find_secondary_label(self, container):
        try:
            children = container.get_children()
            if len(children) == 0:
                return None
            if isinstance(children[0], Gtk.Label):
                return children[1]
            for child in children:
                label = self.find_secondary_label(child)
                if label is not None:
                    return label
        except AttributeError:
            pass
        except IndexError:
            pass
        return None

    def configure_application_window(self, app):
        window_title = "MenuLibre"

        # Initialize the GtkApplicationWindow.
        Gtk.Window.__init__(self, title=window_title, application=app)
        self.set_wmclass(window_title, "MenuLibre")

        # Restore the window properties.
        self.set_title("MenuLibre")
        self.set_icon_name("menulibre")
        self.set_default_size(-1, 500)
        self.set_size_request(-1, 640)

        # Reparent the widgets.
        box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        self.add(box)

        self.toolbar_container = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        box.pack_start(self.toolbar_container, False, False, 0)

        self.infobar_container = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        box.pack_start(self.infobar_container, False, False, 0)

        self.panes = Gtk.Paned.new(Gtk.Orientation.HORIZONTAL)
        box.pack_start(self.panes, True, True, 0)

        # Connect any window-specific events.
        self.connect('key-press-event', self.on_window_keypress_event)
        self.connect('delete-event', self.on_window_delete_event)

    def configure_css(self):
        screen = Gdk.Screen.get_default()
        if screen is None:
            return

        css = """
        #MenulibreSidebarToolbar {
            border-left-width: 0;
            border-right-width: 0;
            border-bottom-width: 0;
            border-radius: 0;
        }
        #MenulibreSidebarScroll.frame {
            border-left-width: 0;
            border-right-width: 0;
        }
        #helpbutton {
            padding: 0;
        }
        #hideybutton {
            padding: 0;
            border-radius: 0;
            min-height: 18px;
            min-width: 18px;
        }
        """

        style_provider = Gtk.CssProvider.new()
        style_provider.load_from_data(bytes(css.encode()))

        Gtk.StyleContext.add_provider_for_screen(
            screen, style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def configure_headerbar(self, add_menu):
        # Configure the Add, Save, Undo, Redo, Revert, Delete widgets.
        headerbar = Headerbar()

        self.add_button = headerbar.add_menu_button(
            "list-add-symbolic", _("Add..."), add_menu)
        self.save_button = headerbar.add_button(
            "document-save-symbolic", _("Save Launcher"))

        self.undo_button, self.redo_button = headerbar.add_buttons([
            ["edit-undo-symbolic", _("Undo")],
            ["edit-redo-symbolic", _("Redo")],
        ])

        self.revert_button = headerbar.add_button(
            "document-revert-symbolic", _("Revert"))
        self.execute_button = headerbar.add_button(
            "media-playback-start-symbolic", _("Test Launcher"))
        self.delete_button = headerbar.add_button(
            "edit-delete-symbolic", _("Delete..."))

        headerbar.add_search(self.search_box)

        self.set_titlebar(headerbar)
        headerbar.show_all()

    def configure_application_actions(self):
        """Configure the GtkActions that are used in the Menulibre
        application."""
        self.actions = {}

        # Add Launcher
        self.actions['add_launcher'] = Gtk.Action(
            name='add_launcher',
            # Translators: Add Launcher action label
            label=_('Add _Launcher…'),
            # Translators: Add Launcher action tooltip
            tooltip=_('Add Launcher…'),
            stock_id=Gtk.STOCK_NEW)

        # Add Directory
        self.actions['add_directory'] = Gtk.Action(
            name='add_directory',
            # Translators: Add Directory action label
            label=_('Add _Directory…'),
            # Translators: Add Directory action tooltip
            tooltip=_('Add Directory…'),
            stock_id=Gtk.STOCK_NEW)

        # Add Separator
        self.actions['add_separator'] = Gtk.Action(
            name='add_separator',
            # Translators: Add Separator action label
            label=_('_Add Separator…'),
            # Translators: Add Separator action tooltip
            tooltip=_('Add Separator…'),
            stock_id=Gtk.STOCK_NEW)

        # Save Launcher
        self.actions['save_launcher'] = Gtk.Action(
            name='save_launcher',
            # Translators: Save Launcher action label
            label=_('_Save'),
            # Translators: Save Launcher action tooltip
            tooltip=_('Save'),
            stock_id=Gtk.STOCK_SAVE)

        # Undo
        self.actions['undo'] = Gtk.Action(
            name='undo',
            # Translators: Undo action label
            label=_('_Undo'),
            # Translators: Undo action tooltip
            tooltip=_('Undo'),
            stock_id=Gtk.STOCK_UNDO)

        # Redo
        self.actions['redo'] = Gtk.Action(
            name='redo',
            # Translators: Redo action label
            label=_('_Redo'),
            # Translators: Redo action tooltip
            tooltip=_('Redo'),
            stock_id=Gtk.STOCK_REDO)

        # Revert
        self.actions['revert'] = Gtk.Action(
            name='revert',
            # Translators: Revert action label
            label=_('_Revert'),
            # Translators: Revert action tooltip
            tooltip=_('Revert'),
            stock_id=Gtk.STOCK_REVERT_TO_SAVED)

        # Execute
        self.actions['execute'] = Gtk.Action(
            name='execute',
            # Translators: Execute action label
            label=_('_Execute'),
            # Translators: Execute action tooltip
            tooltip=_('Execute Launcher'),
            stock_id=Gtk.STOCK_MEDIA_PLAY)

        # Delete
        self.actions['delete'] = Gtk.Action(
            name='delete',
            # Translators: Delete action label
            label=_('_Delete'),
            # Translators: Delete action tooltip
            tooltip=_('Delete'),
            stock_id=Gtk.STOCK_DELETE)

        # Quit
        self.actions['quit'] = Gtk.Action(
            name='quit',
            # Translators: Quit action label
            label=_('_Quit'),
            # Translators: Quit action tooltip
            tooltip=_('Quit'),
            stock_id=Gtk.STOCK_QUIT)

        # Help
        self.actions['help'] = Gtk.Action(
            name='help',
            # Translators: Help action label
            label=_('_Contents'),
            # Translators: Help action tooltip
            tooltip=_('Help'),
            stock_id=Gtk.STOCK_HELP)

        # About
        self.actions['about'] = Gtk.Action(
            name='about',
            # Translators: About action label
            label=_('_About'),
            # Translators: About action tooltip
            tooltip=_('About'),
            stock_id=Gtk.STOCK_ABOUT)

        # Connect the GtkAction events.
        self.actions['add_launcher'].connect('activate',
                                             self.on_add_launcher_cb)
        self.actions['add_directory'].connect('activate',
                                              self.on_add_directory_cb)
        self.actions['add_separator'].connect('activate',
                                              self.on_add_separator_cb)
        self.actions['save_launcher'].connect('activate',
                                              self.on_save_launcher_cb)
        self.actions['undo'].connect('activate', self.on_undo_cb)
        self.actions['redo'].connect('activate', self.on_redo_cb)
        self.actions['revert'].connect('activate', self.on_revert_cb)
        self.actions['execute'].connect('activate', self.on_execute_cb)
        self.actions['delete'].connect('activate', self.on_delete_cb)
        self.actions['quit'].connect('activate', self.on_quit_cb)
        self.actions['help'].connect('activate', self.on_help_cb)
        self.actions['about'].connect('activate', self.on_about_cb)

    def configure_application_bad_desktop_files_infobar(self):
        """Configure InfoBar to alert user to bad desktop files."""
        self.infobar = Gtk.InfoBar.new()
        self.infobar.set_message_type(Gtk.MessageType.WARNING)
        self.infobar.set_no_show_all(True)
        self.infobar_container.add(self.infobar)

        content = self.infobar.get_content_area()

        label = Gtk.Label.new(
            _("Invalid desktop files detected! Please see details."))
        label.show()
        content.add(label)

        self.infobar.add_button(_('Details'), Gtk.ResponseType.YES)

        self.infobar.show()

        # Hook up events
        self.infobar.set_show_close_button(True)
        self.infobar.set_default_response(Gtk.ResponseType.CLOSE)

        self.infobar.connect('response',
                             self.on_bad_desktop_files_infobar_response)

    def configure_menu_restart_infobar(self):
        self.menu_restart_infobar = Gtk.InfoBar.new()
        self.menu_restart_infobar.set_message_type(Gtk.MessageType.WARNING)
        self.menu_restart_infobar.set_no_show_all(True)
        self.infobar_container.add(self.menu_restart_infobar)

        content = self.menu_restart_infobar.get_content_area()

        label = Gtk.Label.new(
            _("Your applications menu may need to be restarted."))
        label.show()
        content.add(label)

        self.menu_restart_infobar.add_button(
            _("Restart menu..."), Gtk.ResponseType.ACCEPT)

        self.menu_restart_infobar.set_show_close_button(True)
        self.menu_restart_infobar.set_default_response(Gtk.ResponseType.CLOSE)

        self.menu_restart_infobar.connect(
            'response', self.on_menu_restart_infobar_response)

    def on_menu_restart_infobar_response(self, infobar, response_id):
        if response_id == Gtk.ResponseType.CLOSE:
            infobar.hide()
        elif response_id == Gtk.ResponseType.ACCEPT:
            self.on_menu_restart_button_activate(infobar)

    def configure_application_toolbar(self, add_menu):
        """Configure the application toolbar."""
        toolbar = Toolbar()
        self.toolbar_container.add(toolbar)

        self.add_button = toolbar.add_menu_button(
            "list-add", _("Add..."), add_menu)

        toolbar.add_separator()

        self.save_button = toolbar.add_button("document-save", _("Save"))

        toolbar.add_separator()

        self.undo_button = toolbar.add_button("edit-undo", _("Undo"))
        self.redo_button = toolbar.add_button("edit-redo", _("Redo"))

        toolbar.add_separator()

        self.revert_button = toolbar.add_button("document-revert", _("Revert"))

        toolbar.add_separator()

        self.execute_button = toolbar.add_button(
            "system-run", _("Test Launcher"))

        toolbar.add_separator()

        self.delete_button = toolbar.add_button("edit-delete", _("Delete..."))

        separator = toolbar.add_separator()
        separator.set_draw(False)
        separator.set_expand(True)

        toolbar.add_search(self.search_box)

        toolbar.show_all()

    def configure_application_treeview(self):
        """Configure the menu-browsing GtkTreeView."""
        self.treeview = MenulibreTreeview.Treeview(self)
        if not self.treeview.loaded:
            self.menu_load_failure()

        self.panes.add(self.treeview)

        self.editor = ApplicationEditor(use_headerbar=self.use_headerbar)
        self.panes.add(self.editor)

        self.treeview.set_search_entry(self.search_box)
        self.search_box.connect('changed', self.on_app_search_changed, True)
        self.treeview.set_can_select_function(self.get_can_select)
        self.treeview.connect("cursor-changed",
                              self.on_apps_browser_cursor_changed)
        self.treeview.connect("add-directory-enabled",
                              self.on_apps_browser_add_directory_enabled)
        self.treeview.connect("requires-menu-reload",
                              self.on_apps_browser_requires_menu_reload)
        self.treeview.reset_cursor()

        self.editor.connect("value-changed", self.on_smart_widget_changed)

    def get_can_select(self):
        if self.save_button.get_sensitive():
            dialog = Dialogs.SaveOnLeaveDialog(self, self.use_headerbar)

            response = dialog.run()
            dialog.destroy()
            # Cancel prevents leaving this launcher.
            if response == Gtk.ResponseType.CANCEL:
                return False
            # Don't Save allows leaving this launcher, deleting 'new'.
            elif response == Gtk.ResponseType.NO:
                filename = self.treeview.get_selected_filename()
                if filename is None:
                    self.delete_launcher()
                    return False
                return True
            # Save and move on.
            else:
                self.save_launcher()
                return True
        else:
            return True

    def activate_action_cb(self, widget, action_name):
        """Activate the specified GtkAction."""
        self.actions[action_name].activate()

# History Signals
    def on_undo_changed(self, history, enabled):
        """Toggle undo functionality when history is changed."""
        self.undo_button.set_sensitive(enabled)

    def on_redo_changed(self, history, enabled):
        """Toggle redo functionality when history is changed."""
        self.redo_button.set_sensitive(enabled)

    def on_revert_changed(self, history, enabled):
        """Toggle revert functionality when history is changed."""
        self.revert_button.set_sensitive(enabled)
        self.save_button.set_sensitive(enabled)
        self.actions['save_launcher'].set_sensitive(enabled)
        self.emit('action-enabled', 'save_launcher', enabled)

# Generic Treeview functions
    def treeview_add(self, treeview, row_data):
        """Append the specified row_data to the treeview."""
        model = treeview.get_model()
        model.append(row_data)

    def treeview_remove(self, treeview):
        """Remove the selected row from the treeview."""
        model, treeiter = treeview.get_selection().get_selected()
        if model is not None and treeiter is not None:
            model.remove(treeiter)

    def treeview_clear(self, treeview):
        """Remove all items from the treeview."""
        model = treeview.get_model()
        model.clear()

    def treeview_get_selected_text(self, treeview, column):
        """Return selected item's text value stored at the given column (text
        is the expected data type)."""

        # Note that the categories treeview is configured to only allow one row
        # to be selected
        model, treeiter = treeview.get_selection().get_selected()
        if model is not None and treeiter is not None:
            return model[treeiter][column]
        else:
            return ''

# Categories

    def on_smart_widget_changed(self, widget, key, value):
        self.set_value(key, value, False)

        if key == 'Filename':
            self.set_editor_filename(value)

    def cleanup_actions(self):
        """Cleanup the Actions treeview. Remove any rows where name or command
        have not been set."""
        self.editor.remove_incomplete_actions()

# Window events
    def on_window_keypress_event(self, widget, event, user_data=None):
        """Handle window keypress events."""
        # Ctrl-F (Find)
        if check_keypress(event, ['Control', 'f']):
            self.search_box.grab_focus()
            return True
        # Ctrl-S (Save)
        if check_keypress(event, ['Control', 's']):
            self.actions['save_launcher'].activate()
            return True
        # Ctrl-Q (Quit)
        if check_keypress(event, ['Control', 'q']):
            self.actions['quit'].activate()
            return True
        return False

    def on_window_delete_event(self, widget, event):
        """Save changes on close."""
        if self.save_button.get_sensitive():
            # Unsaved changes
            dialog = Dialogs.SaveOnCloseDialog(self, self.use_headerbar)
            response = dialog.run()
            dialog.destroy()
            # Cancel prevents the application from closing.
            if response == Gtk.ResponseType.CANCEL:
                return True
            # Don't Save allows the application to close.
            elif response == Gtk.ResponseType.NO:
                return False
            # Save and close.
            else:
                self.save_launcher()
                return False
        return False


# Applications Treeview

    def on_apps_browser_requires_menu_reload(self, widget, required):
        self.menu_restart_infobar.show()

    def on_menu_restart_button_activate(self, widget):
        processes = getProcessList()
        if "mate-panel" in processes:
            cmd = ["mate-panel", "--replace"]
        elif "xfce4-panel" in processes:
            cmd = ["xfce4-panel", "--restart"]
        elif "unity" in processes:
            cmd = ["unity", "--replace"]
        else:
            self.menu_unable_to_restart_dialog()
            return

        self.menu_restart_dialog(cmd)

    def menu_restart_dialog(self, cmd):
        user_cmd = " ".join(cmd)

        primary = _("Menu restart required")
        secondary = _(
            "MenuLibre can do this automatically. "
            "Do you want to run the following command?\n\n%s") % user_cmd

        dialog = Gtk.MessageDialog(
            transient_for=self.get_toplevel(),  # type: ignore
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.YES_NO,
            text=primary,
            use_header_bar=False)
        dialog.format_secondary_markup(secondary)
        response = dialog.run()
        if response == Gtk.ResponseType.YES:
            subprocess.Popen(cmd)
            self.menu_restart_infobar.hide()
        dialog.destroy()

    def menu_unable_to_restart_dialog(self):
        primary = _("Menu restart required")
        secondary = _("However, MenuLibre cannot determine how to do this "
                      "automatically on your system. "
                      "Please log out for you menu to update completely.")

        dialog = Gtk.MessageDialog(
            transient_for=self.get_toplevel(),  # type: ignore
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.CLOSE,
            text=primary,
            use_header_bar=False)
        dialog.format_secondary_markup(secondary)
        dialog.run()
        self.menu_restart_infobar.hide()
        dialog.destroy()

    def on_apps_browser_add_directory_enabled(self, widget, enabled):
        """Update the Add Directory menu item when the selected row is
        changed."""
        # Always allow creating sub directories
        enabled = True

        self.actions['add_directory'].set_sensitive(enabled)
        self.emit('action-enabled', 'add_directory', enabled)

    def on_apps_browser_cursor_changed(self, widget, value):  # noqa
        """Update the editor frame when the selected row is changed."""
        missing = False

        # Clear history
        self.history.clear()

        # Hide the Name and Comment editors
        self.editor.cancel()

        # Prevent updates to history.
        self.history.block()

        # Clear the individual entries.
        for key in ['Exec', 'Path', 'Terminal', 'StartupNotify',
                    'NoDisplay', 'GenericName', 'TryExec',
                    'OnlyShowIn', 'NotShowIn', 'MimeType',
                    'Keywords', 'StartupWMClass', 'Implements', 'Categories',
                    'Hidden', 'DBusActivatable', 'PrefersNonDefaultGPU',
                    'X-GNOME-UsesNotifications']:
            self.set_value(key, None)

        # Clear the Actions and Icon.
        self.set_value('Actions', None, store=True)
        self.set_value('Icon', None, store=True)

        model, row_data = self.treeview.get_selected_row_data()
        if row_data is None:
            return

        item_type = row_data[MenuEditor.COL_TYPE]

        # If the selected row is a separator, hide the editor.
        if item_type == MenuItemTypes.SEPARATOR:  # type: ignore
            self.editor.hide()
            # Translators: Separator menu item
            self.set_value('Name', _("Separator"), store=True)
            self.set_value('Comment', "", store=True)
            self.set_value('Filename', None, store=True)
            self.set_value('Type', 'Separator', store=True)
            self.execute_button.set_sensitive(False)

        # Otherwise, show the editor and update the values.
        else:
            filename = self.treeview.get_selected_filename()
            new_launcher = filename is None

            # Check if this file still exists
            if (not new_launcher) and (not os.path.isfile(filename)):
                # If it does not, try to fallback...
                basename = getRelativeName(filename)
                filename = util.getSystemLauncherPath(basename)
                if filename is not None:
                    row_data[MenuEditor.COL_FILENAME] = filename
                    self.treeview.update_launcher_instances(filename, row_data)

            if new_launcher or (filename is not None):
                self.editor.show()
                name = row_data[MenuEditor.COL_NAME]
                comment = row_data[MenuEditor.COL_COMMENT]

                self.set_value('Icon',
                               row_data[MenuEditor.COL_ICON_NAME],
                               store=True)
                self.set_value('Name', name, store=True)
                self.set_value('Comment', comment, store=True)
                self.set_value('Filename', filename, store=True)

                if item_type == MenuItemTypes.APPLICATION:  # type: ignore
                    self.editor.show_all()
                    entry = MenulibreXdg.MenulibreDesktopEntry(filename)
                    for key in getRelatedKeys(item_type, key_only=True):
                        if key in ['Actions', 'Comment', 'Filename', 'Icon',
                                   'Name']:
                            continue
                        self.set_value(key, entry[key], store=True)
                    self.set_value('Actions', entry.get_actions(),
                                   store=True)
                    self.set_value('Type', 'Application')
                    self.execute_button.set_sensitive(True)
                else:
                    entry = MenulibreXdg.MenulibreDesktopEntry(filename)
                    for key in getRelatedKeys(item_type, key_only=True):
                        if key in ['Comment', 'Filename', 'Icon', 'Name']:
                            continue
                        self.set_value(key, entry[key], store=True)
                    self.set_value('Type', 'Directory')
                    self.execute_button.set_sensitive(False)

            else:
                # Display a dialog saying this item is missing
                dialog = Dialogs.LauncherRemovedDialog(
                    self, self.use_headerbar)
                dialog.run()
                dialog.destroy()
                # Mark this item as missing to delete it later.
                missing = True

        # Re-enable updates to history.
        self.history.unblock()

        if self.treeview.get_parent()[1] is None:
            self.treeview.set_sortable(False)
            move_up_enabled = not self.treeview.is_first()
            move_down_enabled = not self.treeview.is_last()
        else:
            self.treeview.set_sortable(True)
            if item_type in [MenuItemTypes.APPLICATION,  # type: ignore
                             MenuItemTypes.LINK,  # type: ignore
                             MenuItemTypes.SEPARATOR]:  # type: ignore
                move_up_enabled = True
                move_down_enabled = True
            else:
                move_up_enabled = not self.treeview.is_first()
                move_down_enabled = not self.treeview.is_last()

        self.treeview.set_move_up_enabled(move_up_enabled)
        self.treeview.set_move_down_enabled(move_down_enabled)

        # Remove this item if it happens to be gone.
        if missing:
            self.delete_launcher()

    def on_app_search_changed(self, widget, expand=False):
        """Update search results when query text is modified."""
        query = widget.get_text()

        # If blank query...
        if len(query) == 0:
            # Remove the clear button.
            widget.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY,
                                           None)

            # If the model is a filter, we want to remove the filter.
            self.treeview.set_searchable(False, expand)

            # Enable add functionality
            for name in ['add_launcher', 'add_directory', 'add_separator',
                         'add_button']:
                if name in self.action_items:
                    for widget in self.action_items[name]:
                        widget.set_sensitive(True)
                if name in self.actions:
                    self.actions[name].set_sensitive(True)
                    self.emit('action-enabled', name, True)

            # Enable deletion (LP: #1751616)
            # self.delete_button.set_sensitive(True)
            # self.delete_button.set_tooltip_text(_("Delete..."))

        # If the entry has a query...
        else:
            # Show the clear button.
            widget.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY,
                                           'edit-clear-symbolic')

            self.treeview.set_searchable(True)

            # Disable add functionality
            for name in ['add_launcher', 'add_directory', 'add_separator',
                         'add_button']:
                if name in self.action_items:
                    for widget in self.action_items[name]:
                        widget.set_sensitive(False)
                if name in self.actions:
                    self.actions[name].set_sensitive(False)
                    self.emit('action-enabled', name, False)

            # Rerun the filter.
            self.treeview.search(self.search_box.get_text())

            # Disable deletion (LP: #1751616)
            self.delete_button.set_sensitive(False)
            self.delete_button.set_tooltip_text(_("Delete..."))

    def on_search_cleared(self, widget, event, user_data=None):
        """Generic search cleared callback function."""
        widget.set_text("")

# Setters and Getters
    def set_editor_filename(self, filename):
        """Set the editor filename."""
        # Since the filename has changed, check if it is now writable...
        if filename is None or os.access(filename, os.W_OK):
            self.delete_button.set_sensitive(True)
            self.delete_button.set_tooltip_text(_("Delete..."))
        else:
            self.delete_button.set_sensitive(False)
            self.delete_button.set_tooltip_text(
                # Translators: This error is displayed when the user does not
                # have sufficient file system permissions to delete the
                # selected file.
                _("You do not have permission to delete this file."))

        # Disable deletion if we're in search mode (LP: #1751616)
        if self.search_box.get_text() != "":
            self.delete_button.set_sensitive(False)
            self.delete_button.set_tooltip_text(
                _("You cannot delete this file while a search is active."))

    def get_inner_value(self, key):
        """Get the value stored for key."""
        try:
            return self.values[key]
        except:  # noqa
            return None

    def set_inner_value(self, key, value):
        """Set the value stored for key."""
        self.values[key] = value

    """
    Quoting must be done by enclosing the argument between double quotes and escaping
    the double quote character, backtick character ("`"), dollar sign ("$") and
    backslash character ("\") by preceding it with an additional backslash character.
    Implementations must undo quoting before expanding field codes and before passing
    the argument to the executable program. Reserved characters are space (" "),
    tab, newline, double quote, single quote ("'"), backslash character ("\"),
    greater-than sign (">"), less-than sign ("<"), tilde ("~"), vertical bar ("|"),
    ampersand ("&"), semicolon (";"), dollar sign ("$"), asterisk ("*"),
    question mark ("?"), hash mark ("#"), parenthesis ("(") and (")") and backtick
    character ("`").

    Note that the general escape rule for values of type string states that the backslash
    character can be escaped as ("\\") as well and that this escape rule is applied before
    the quoting rule. As such, to unambiguously represent a literal backslash character in
    a quoted argument in a desktop entry file requires the use of four successive backslash
    characters ("\\\\"). Likewise, a literal dollar sign in a quoted argument in a desktop
    entry file is unambiguously represented with ("\\$").

    A number of special field codes have been defined which will be expanded by the file
    manager or program launcher when encountered in the command line. Field codes consist
    of the percentage character ("%") followed by an alpha character. Literal percentage
    characters must be escaped as %%. Deprecated field codes should be removed from the
    command line and ignored. Field codes are expanded only once, the string that is used
    to replace the field code should not be checked for field codes itself.

    Field codes must not be used inside a quoted argument, the result of field code expansion
    inside a quoted argument is undefined. The %F and %U field codes may only be used as an
    argument on their own.
    """

    def set_value(self, key, value, adjust_widget=True, store=False):  # noqa
        """Set the DesktopSpec key, value pair in the editor."""
        if store:
            self.history.store(key, value)
        if not self.get_inner_value(key) == value:
            self.history.append(key, self.get_inner_value(key), value)
            self.set_inner_value(key, value)
        if not adjust_widget:
            return
        else:
            self.editor.set_value(key, value)
            if key == 'Filename':
                self.set_editor_filename(value)

    def get_value(self, key):  # noqa
        """Return the value stored for the specified key."""
        return self.editor.get_value(key)

# Action Functions
    def add_launcher(self):
        """Add Launcher callback function."""
        # Translators: Placeholder text for a newly created launcher.
        name = _("New Launcher")
        display_name = name
        # Translators: Placeholder text for a newly created launcher's
        # description.
        comment = ""
        categories = ""
        item_type = MenuItemTypes.APPLICATION  # type: ignore
        icon_name = "application-x-executable"
        icon = Gio.ThemedIcon.new(icon_name)
        filename = None
        executable = ""
        new_row_data = [name, display_name, comment, executable, categories, item_type, icon,
                        icon_name, filename, True]

        model, parent_data = self.treeview.get_parent_row_data()
        model, row_data = self.treeview.get_selected_row_data()

        # Exit early if no row is selected (LP #1556664)
        if not row_data:
            return

        # Add to the treeview on the current level or as a child of a selected
        # directory
        dir_selected = row_data[MenuEditor.COL_TYPE] == MenuItemTypes.DIRECTORY  # type: ignore
        if dir_selected:
            self.treeview.add_child(new_row_data)
        else:
            self.treeview.append(new_row_data)

        if parent_data is not None and not dir_selected:
            # A parent item has been found, and the current selection is not a
            # directory, so the resulting item will be placed at the current level
            # fetch the parent's categories
            parent_directory = parent_data[MenuEditor.COL_FILENAME]

        elif parent_data is not None and dir_selected:
            # A directory lower than the top-level has been selected - the
            # launcher will be added into it (e.g. as the first item),
            # therefore it essentially has a parent of the current selection
            parent_directory = row_data[MenuEditor.COL_FILENAME]

        else:
            # Parent was not found, this is a toplevel category
            parent_directory = None

        self.editor.clear_categories()
        self.editor.insert_required_categories(parent_directory)

        self.actions['save_launcher'].set_sensitive(True)
        self.emit('action-enabled', 'save_launcher', True)
        self.save_button.set_sensitive(True)

        self.editor.take_focus()

    def add_directory(self):
        """Add Directory callback function."""
        # Translators: Placeholder text for a newly created directory.
        name = _("New Directory")
        display_name = name
        # Translators: Placeholder text for a newly created directory's
        # description.
        comment = ""
        categories = ""
        item_type = MenuItemTypes.DIRECTORY  # type: ignore
        icon_name = "folder"
        icon = Gio.ThemedIcon.new(icon_name)
        filename = None
        executable = ""
        row_data = [name, display_name, comment, executable, categories, item_type, icon,
                    icon_name, filename, True, True]

        self.treeview.append(row_data)

        self.actions['save_launcher'].set_sensitive(True)
        self.emit('action-enabled', 'save_launcher', True)
        self.save_button.set_sensitive(True)

        self.editor.take_focus()

    def add_separator(self):
        """Add Separator callback function."""
        name = _("Separator")
        display_name = name
        # Translators: Separator menu item
        tooltip = _("Separator")
        categories = ""
        filename = None
        icon_name = "content-loading-symbolic"
        icon = Gio.ThemedIcon.new(icon_name)
        item_type = MenuItemTypes.SEPARATOR  # type: ignore
        filename = None
        executable = ""
        row_data = [name, display_name, tooltip, executable, categories, item_type, icon,
                    icon_name, filename, False, False]

        self.treeview.append(row_data)

        self.save_button.set_sensitive(False)

        self.treeview.update_menus()

    def list_str_to_list(self, value):
        if isinstance(value, list):
            return value
        values = []
        for value in value.replace(",", ";").split(";"):
            value = value.strip()
            if len(value) > 0:
                values.append(value)
        return values

    def write_launcher(self, filename):  # noqa
        keyfile = GLib.KeyFile.new()

        for key, ktype, required in getRelatedKeys(self.get_value("Type")):
            if key == "Version":
                keyfile.set_string("Desktop Entry", "Version", "1.1")
                continue

            if key == "Actions":
                action_list = []
                for show, name, displayed, command in \
                        self.editor.get_actions():
                    group_name = "Desktop Action %s" % name
                    keyfile.set_string(group_name, "Name", displayed)
                    keyfile.set_string(group_name, "Exec", command)
                    if show:
                        action_list.append(name)
                keyfile.set_string_list("Desktop Entry", key, action_list)
                continue

            value = self.get_value(key)
            if ktype == str:
                if len(value) > 0:  # type: ignore
                    keyfile.set_string("Desktop Entry", key, value)  # type: ignore
            if ktype == float:
                if value != 0:
                    keyfile.set_double("Desktop Entry", key, value)  # type: ignore
            if ktype == bool:
                if value is not False:
                    keyfile.set_boolean("Desktop Entry", key, value)  # type: ignore
            if ktype == list:
                value = self.list_str_to_list(value)
                if len(value) > 0:
                    keyfile.set_string_list("Desktop Entry", key, value)

        try:
            if not keyfile.save_to_file(filename):
                return False
        except GLib.Error:
            return False

        return True

    def save_launcher(self, temp=False):  # noqa
        """Save the current launcher details, remove from the current directory
        if it no longer has the required category."""

        item_type = None
        original_filename = None

        if temp:
            filename = tempfile.mkstemp('.desktop', 'menulibre-')[1]
        else:
            # Get the filename to be used.
            original_filename = self.get_value('Filename')
            item_type = self.get_value('Type')
            name = self.get_value('Name')
            filename = util.getSaveFilename(name, original_filename, item_type)
        logger.debug("Saving launcher as \"%s\"" % filename)

        if not temp:
            model, row_data = self.treeview.get_selected_row_data()
            if row_data is None:
                dlg = Dialogs.SaveErrorDialog(self, filename, self.use_headerbar)
                dlg.run()
                return False

            item_type = row_data[MenuEditor.COL_TYPE]

            model, parent_data = self.treeview.get_parent_row_data()

            # Make sure required categories are in place - this is useful for
            # when a user moves a launcher from its original location to a new
            # directory - without the category associated with the new
            # directory (and no force-include), the launcher would not
            # otherwise show
            if parent_data is not None:
                # Parent was found, take its categories.
                parent_directory = parent_data[MenuEditor.COL_FILENAME]
            else:
                # Parent was not found, this is a toplevel category
                parent_directory = None

            self.editor.insert_required_categories(parent_directory)

            # Cleanup invalid entries and reorder the Categories and Actions
            self.cleanup_actions()

        if not self.write_launcher(filename):
            dlg = Dialogs.SaveErrorDialog(self, filename, self.use_headerbar)
            dlg.run()
            return False

        if temp:
            return filename

        # Install the new item in its directory...
        self.treeview.xdg_menu_install(filename)

        # Set the editor to the new filename.
        self.set_value('Filename', filename)

        # Update the selected iter with the new details.
        name = self.get_value('Name')
        comment = self.get_value('Comment')
        executable = self.get_value('Exec')
        categories = self.get_value('Categories')
        icon_name = self.get_value('Icon')
        hidden = self.get_value('Hidden') or self.get_value('NoDisplay')
        self.treeview.update_selected(
            name,
            comment,
            executable,
            categories,
            item_type,
            icon_name,
            filename,
            not hidden)
        self.history.clear()

        # Update all instances
        model, row_data = self.treeview.get_selected_row_data()
        self.treeview.update_launcher_instances(original_filename, row_data)

        self.treeview.update_menus()

        # Check and make sure that the launcher has been added to/removed from
        # directories that its category configuration dictates - this is not
        # deleting the launcher but removing it from various places in the UI
        self.update_launcher_category_dirs()

        if filename.endswith(".directory"):
            processes = getProcessList()
            if "mate-panel" in processes:
                self.menu_restart_infobar.show()

        return True

    def update_launcher_categories(self, remove, add):  # noqa
        original_filename = self.get_value('Filename')
        if original_filename is None or not os.path.isfile(original_filename):
            return
        item_type = self.get_value('Type')
        name = self.get_value('Name')
        save_filename = util.getSaveFilename(name, original_filename,
                                             item_type, force_update=True)
        logger.debug("Saving launcher as \"%s\"" % save_filename)

        # Get the original contents
        keyfile = GLib.KeyFile.new()
        keyfile.load_from_file(original_filename,  # type: ignore
                               GLib.KeyFileFlags.NONE)

        try:
            categories = keyfile.get_string_list("Desktop Entry", "Categories")
        except GLib.Error:
            categories = None

        if categories is None:
            categories = []

        # Remove the old required categories
        for category in remove:
            if category in categories:
                categories.remove(category)

        # Add the new required categories
        for category in add:
            if category not in categories:
                categories.append(category)

        # Remove empty categories
        for category in categories:
            if category.strip() == "":
                try:
                    categories.remove(category)
                except:  # noqa
                    pass

        categories.sort()

        # Commit the changes to a new file
        keyfile.set_string_list("Desktop Entry", "Categories", categories)
        keyfile.save_to_file(save_filename)

        # Set the editor to the new filename.
        self.set_value('Filename', save_filename)

        # Update all instances
        model, row_data = self.treeview.get_selected_row_data()
        if row_data is None:
            return
        row_data[MenuEditor.COL_CATEGORIES] = ';'.join(categories)
        row_data[MenuEditor.COL_FILENAME] = save_filename
        self.treeview.update_launcher_instances(original_filename, row_data)

    def update_launcher_category_dirs(self):  # noqa
        """Make sure launcher is present or absent from in all top-level
        directories that its categories dictate."""

        # Prior to menulibre being restarted, addition of a category does not
        # result in the launcher being added to or removed from the relevant
        # top-level directories - making sure this happens

        # Fetching model and launcher information - removing empty category
        # at end of category split
        # Note that a user can remove all categories now if they want, which
        # would naturally remove the launcher from all top-level directories -
        # alacarte doesn't save any categories by default with a new launcher,
        # however to reach this point, any required categories (minus those the
        # user has explicitly deleted) will be added, so this shouldn't be a
        # problem
        model, row_data = self.treeview.get_selected_row_data()
        if row_data is None:
            return
        if row_data[MenuEditor.COL_CATEGORIES]:
            categories = row_data[MenuEditor.COL_CATEGORIES].split(';')[:-1]
        else:
            categories = []
        filename = row_data[MenuEditor.COL_FILENAME]

        required_category_directories = set()

        # Obtaining a dictionary of iters to launcher instances in top-level
        # directories
        launcher_instances = self.treeview._get_launcher_instances(filename)
        launchers_in_top_level_dirs = {}
        for instance in launcher_instances:

            # Make sure the launcher isn't top-level and is in a directory.
            # Must pass a model otherwise it gets the current selection iter
            # regardless...
            _, parent = self.treeview.get_parent(model, instance)
            if (parent is not None and
                    model[parent][MenuEditor.COL_TYPE] == MenuItemTypes.DIRECTORY):  # type: ignore

                # Any direct parents are required directories.
                required_category_directories.add(
                    model[parent][MenuEditor.COL_NAME])

                # Adding if the directory returned is top level
                _, parent_parent = self.treeview.get_parent(model, parent)
                if parent_parent is None:
                    launchers_in_top_level_dirs[model[parent][
                        MenuEditor.COL_NAME]] = instance

        # Obtaining a lookup of top-level directories -> iters
        top_level_dirs = {}
        for row in model:
            if row[MenuEditor.COL_TYPE] == MenuItemTypes.DIRECTORY:  # type: ignore
                top_level_dirs[row[MenuEditor.COL_NAME]
                               ] = model.get_iter(row.path)

        # Looping through all set categories - category specified is at maximum
        # detail level, this needs to be converted to the parent group name,
        # and this needs to be converted into the directory name as it would
        # appear in the menu
        for category in categories:
            if category not in category_lookup.keys():
                continue

            category_group = category_lookup[category]
            directory_name = util.getDirectoryNameFromCategory(category_group)

            # Adding to directories the launcher should be in
            if directory_name not in launchers_in_top_level_dirs:
                if directory_name in top_level_dirs.keys():
                    treeiter = self.treeview.add_child(
                        row_data, top_level_dirs[directory_name], model, False)
                    launchers_in_top_level_dirs[directory_name] = treeiter

            # Building set of required category directories to detect
            # superfluous ones later
            if directory_name not in required_category_directories:
                required_category_directories.add(directory_name)

        # Removing launcher from directories it should no longer be in
        superfluous_dirs = (set(launchers_in_top_level_dirs.keys())
                            - required_category_directories)
        _, parent_data = self.treeview.get_parent_row_data()

        for directory_name in superfluous_dirs:

            # Removing selected launcher from the UI if it is in the current
            # directory, otherwise just from the model
            if parent_data is not None and directory_name == parent_data[MenuEditor.COL_NAME]:
                self.treeview.remove_selected(True)

            else:
                self.treeview.remove_iter(
                    model, launchers_in_top_level_dirs[directory_name])

    def delete_separator(self):
        """Remove a separator row from the treeview, update the menu files."""
        self.treeview.remove_selected()

    def delete_launcher(self):
        """Delete the selected launcher."""
        self.treeview.remove_selected()
        self.history.clear()

    def restore_launcher(self):
        """Revert the current launcher."""
        values = self.history.restore()

        # Clear the history
        self.history.clear()

        # Block updates
        self.history.block()

        for key in list(values.keys()):
            self.set_value(key, values[key], store=True)

        # Unblock updates
        self.history.unblock()

# Callbacks
    def on_add_launcher_cb(self, widget):
        """Add Launcher callback function."""
        self.add_launcher()

    def on_add_directory_cb(self, widget):
        """Add Directory callback function."""
        self.add_directory()

    def on_add_separator_cb(self, widget):
        """Add Separator callback function."""
        self.add_separator()

    def on_save_launcher_cb(self, widget):
        """Save Launcher callback function."""
        self.editor.commit()
        self.save_launcher()

    def on_undo_cb(self, widget):
        """Undo callback function."""
        key, value = self.history.undo()
        self.history.block()
        self.set_value(key, value)
        self.history.unblock()

    def on_redo_cb(self, widget):
        """Redo callback function."""
        key, value = self.history.redo()
        self.history.block()
        self.set_value(key, value)
        self.history.unblock()

    def on_revert_cb(self, widget):
        """Revert callback function."""
        dialog = Dialogs.RevertDialog(self, self.use_headerbar)
        if dialog.run() == Gtk.ResponseType.OK:
            self.restore_launcher()
        dialog.destroy()

    def find_in_path(self, command):
        if os.path.exists(os.path.abspath(command)):
            return os.path.abspath(command)
        for path in os.environ["PATH"].split(os.pathsep):
            if os.path.exists(os.path.join(path, command)):
                return os.path.join(path, command)
        return False

    def find_command_in_string(self, command):
        for piece in shlex.split(command):
            if "=" not in piece:
                return piece
        return False

    def on_execute_cb(self, widget):
        """Execute callback function."""
        self.editor.commit()
        filename = self.save_launcher(True)

        if not filename:
            return

        entry = MenulibreXdg.MenulibreDesktopEntry(filename)
        command = self.find_command_in_string(entry["Exec"])

        if self.find_in_path(command):
            subprocess.Popen(["xdg-open", filename])  # type: ignore
            GObject.timeout_add(2000, self.on_execute_timeout, filename)
        else:
            os.remove(filename)  # type: ignore
            dlg = Dialogs.NotFoundInPathDialog(
                self, command, self.use_headerbar)
            dlg.run()

    def on_execute_timeout(self, filename):
        os.remove(filename)

    def on_delete_cb(self, widget):
        """Delete callback function."""
        model, row_data = self.treeview.get_selected_row_data()
        if row_data is None:
            return

        name = row_data[MenuEditor.COL_NAME]
        item_type = row_data[MenuEditor.COL_TYPE]

        # Prepare the strings
        if item_type == MenuItemTypes.SEPARATOR:  # type: ignore
            # Translators: Confirmation dialog to delete the selected
            # separator.
            question = _("Are you sure you want to delete this separator?")
            delete_func = self.delete_separator
        else:
            # Translators: Confirmation dialog to delete the selected launcher.
            question = _("Are you sure you want to delete \"%s\"?") % name
            delete_func = self.delete_launcher

        dialog = Dialogs.DeleteDialog(self, question, self.use_headerbar)

        # Run
        if dialog.run() == Gtk.ResponseType.OK:
            delete_func()

        dialog.destroy()

    def on_quit_cb(self, widget):
        """Quit callback function.  Send the quit signal to the parent
        GtkApplication instance."""
        self.emit('quit', True)

    def on_help_cb(self, widget):
        """Help callback function.  Send the help signal to the parent
        GtkApplication instance."""
        self.emit('help', True)

    def on_about_cb(self, widget):
        """About callback function.  Send the about signal to the parent
        GtkApplication instance."""
        self.emit('about', True)

    def on_bad_desktop_files_infobar_response(self, infobar, response_id):
        """Bad desktop files infobar callback function to request the bad
        desktop files report if desired."""

        # Dealing with request for details
        if response_id == Gtk.ResponseType.YES:
            self.bad_desktop_files_report_dialog()

        # All response types should result in the infobar being hidden
        infobar.hide()

    def bad_desktop_files_report_dialog(self):
        """Generate and display details of bad desktop files, or report
        successful parsing."""

        log_dialog = ParsingErrorsDialog.ParsingErrorsDialog(
            parent=self, use_header_bar=self.use_headerbar
        )

        # Building up a list of all known failures associated with the bad
        # desktop files
        for desktop_file in self.bad_desktop_files:
            log_dialog.add_item(desktop_file,
                                util.validate_desktop_file(desktop_file))

        log_dialog.run()
        log_dialog.destroy()


class Application(Gtk.Application):
    """Menulibre GtkApplication"""

    def __init__(self):
        """Initialize the GtkApplication."""
        Gtk.Application.__init__(self)
        self.use_headerbar = False
        self.use_toolbar = False

        self.settings_file = os.path.expanduser("~/.config/menulibre.cfg")

    def set_use_headerbar(self, preference):
        try:
            settings = GLib.KeyFile.new()
            settings.set_boolean("menulibre", "UseHeaderbar", preference)
            settings.save_to_file(self.settings_file)
        except:  # noqa
            pass

    def get_use_headerbar(self):
        if not os.path.exists(self.settings_file):
            return None
        try:
            settings = GLib.KeyFile.new()
            settings.load_from_file(self.settings_file, GLib.KeyFileFlags.NONE)
            return settings.get_boolean("menulibre", "UseHeaderbar")
        except:  # noqa
            return None

    def do_activate(self):
        """Handle GtkApplication do_activate."""
        if self.use_toolbar:
            headerbar = False
            self.set_use_headerbar(False)
        elif self.use_headerbar:
            headerbar = True
            self.set_use_headerbar(True)
        elif self.get_use_headerbar() is not None:
            headerbar = self.get_use_headerbar()
        elif current_desktop in ["budgie", "gnome", "pantheon"]:
            headerbar = True
        else:
            headerbar = False

        self.win = MenulibreWindow(self, headerbar)  # type: ignore
        self.win.show()

        self.win.connect('about', self.about_cb)
        self.win.connect('help', self.help_cb)
        self.win.connect('quit', self.quit_cb)

    def do_startup(self):
        """Handle GtkApplication do_startup."""
        Gtk.Application.do_startup(self)

        # 'Sections' without labels result in a separator separating functional
        # groups of menu items
        self.menu = Gio.Menu()
        section_1_menu = Gio.Menu()
        # Translators: Menu item to open the Parsing Errors dialog.
        section_1_menu.append(_("Parsing Error Log"),
                              "app.bad_files")
        self.menu.append_section(None, section_1_menu)

        section_2_menu = Gio.Menu()
        section_2_menu.append(_("Help"), "app.help")
        section_2_menu.append(_("About"), "app.about")
        section_2_menu.append(_("Quit"), "app.quit")
        self.menu.append_section(None, section_2_menu)

        self.set_app_menu(self.menu)

        # Bad desktop files detection on demand
        bad_files_action = Gio.SimpleAction.new("bad_files", None)
        bad_files_action.connect("activate", self.bad_files_cb)
        self.add_action(bad_files_action)

        help_action = Gio.SimpleAction.new("help", None)
        help_action.connect("activate", self.help_cb)
        self.add_action(help_action)

        about_action = Gio.SimpleAction.new("about", None)
        about_action.connect("activate", self.about_cb)
        self.add_action(about_action)

        quit_action = Gio.SimpleAction.new("quit", None)
        quit_action.connect("activate", self.quit_cb)
        self.add_action(quit_action)

        add_launcher_action = Gio.SimpleAction.new("add_launcher", None)
        add_launcher_action.connect("activate", self.action_cb, "add_launcher")
        self.add_action(add_launcher_action)

        add_directory_action = Gio.SimpleAction.new("add_directory", None)
        add_directory_action.connect(
            "activate", self.action_cb, "add_directory")
        self.add_action(add_directory_action)

        add_separator_action = Gio.SimpleAction.new("add_separator", None)
        add_separator_action.connect(
            "activate", self.action_cb, "add_separator")
        self.add_action(add_separator_action)

    def bad_files_cb(self, widget, data=None):
        """Bad desktop files detection callback function."""

        # Determining paths of bad desktop files GMenu can't load, on demand
        # This state is normally tracked with the MenulibreWindow, so not
        # keeping it in this application object. By the time this is called,
        # self.win is valid
        self.win.bad_desktop_files = util.determine_bad_desktop_files()
        self.win.bad_desktop_files_report_dialog()

    def help_cb(self, widget, data=None):
        """Help callback function."""
        dialog = Dialogs.HelpDialog(self.win, self.win.use_headerbar)
        dialog.show()

    def about_cb(self, widget, data=None):
        """About callback function.  Display the AboutDialog."""
        dialog = Dialogs.AboutDialog(self.win, self.win.use_headerbar)
        dialog.show()

    def quit_cb(self, widget, data=None):
        """Signal handler for closing the MenulibreWindow."""
        self.quit()

    def action_cb(self, widget, data=None, action_name=None):
        self.win.activate_action_cb(None, action_name)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#   Copyright (C) 2016-2018 OmegaPhil <OmegaPhil@startmail.com>
#
#   Portions of this file are adapted from Alacarte Menu Editor,
#   Copyright (C) 2006 Travis Watkins, Heinrich Wendel
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import locale
import os
import xml.dom.minidom
import xml.parsers.expat
from locale import gettext as _

import logging
logger = logging.getLogger('menulibre')  # noqa

import gi
gi.require_version('GMenu', '3.0')  # noqa
from gi.repository import Gio, GMenu, Gtk  # type: ignore

from . import util
from .util import MenuItemTypes, escapeText, mapDesktopEnvironmentDirectories, unmapDesktopEnvironmentDirectories

locale.textdomain('menulibre')

icon_theme = Gtk.IconTheme.get_default()


def get_icon_theme_name_from_settings():
    try:
        settings = Gtk.Settings.get_default()
        if settings is None:
            return None
        theme_name = settings.get_property("gtk-icon-theme-name")
        if theme_name is not None and len(theme_name) > 0:
            return theme_name
    except BaseException:
        pass
    return None


def get_icon_theme_name_from_icons():
    try:
        lookup = icon_theme.lookup_icon("folder", 16, 0)  # type: ignore
        if lookup is None:
            return None
        filename = lookup.get_filename()
        if filename is None:
            return None
        filename = os.path.realpath(filename)
        path = filename.split('/')
        for i in range(len(path) - 1):
            index = os.path.join('/'.join(path[0:-i]), 'index.theme')
            if os.path.exists(index):
                return path[-i - 1]
    except BaseException:
        pass
    return None


def get_icon_theme_name():
    theme_name = get_icon_theme_name_from_settings()
    if theme_name is not None:
        return theme_name

    theme_name = get_icon_theme_name_from_icons()
    if theme_name is not None:
        return theme_name

    return "Adwaita"


icon_theme_name = get_icon_theme_name()


menu_name = ""


COL_NAME = 0
COL_DISPLAY_NAME = 1
COL_COMMENT = 2
COL_EXEC = 3
COL_CATEGORIES = 4
COL_TYPE = 5
COL_G_ICON = 6
COL_ICON_NAME = 7
COL_FILENAME = 8
COL_EXPAND = 9
COL_SHOW = 10


def get_default_menu():
    """Return the filename of the default application menu."""
    prefixes = [
        util.getDefaultMenuPrefix(),
        ''
    ]
    user_basedir = util.getUserMenusDirectory()
    for prefix in prefixes:
        filename = '%s%s' % (prefix, 'applications.menu')
        user_dir = os.path.join(user_basedir, filename)
        if os.path.exists(user_dir):
            return filename
        system_dir = util.getSystemMenuPath(filename)
        if system_dir:
            return filename
    return None


def menu_to_treestore(treestore, parent, menu_items):
    """Convert the Alacarte menu to a standard treestore."""
    for item in menu_items:
        item_type = item[0]
        if item_type == MenuItemTypes.SEPARATOR:  # type: ignore
            executable = ""
            name = _("Separator")
            displayed_name = name
            # Translators: Separator menu item
            tooltip = _("Separator")
            categories = ""
            filename = None
            icon_name = "content-loading-symbolic"
            icon = Gio.ThemedIcon.new(icon_name)
            item_type = MenuItemTypes.SEPARATOR  # type: ignore
            show = False
        else:
            executable = item[2]['executable']
            name = item[2]['display_name']
            displayed_name = escapeText(name)
            show = item[2]['show']
            tooltip = item[2]['comment']
            categories = item[2]['categories']
            icon = item[2]['icon']
            filename = item[2]['filename']
            icon_name = item[2]['icon_name']

        treeiter = treestore.append(
            parent, [name, displayed_name, tooltip, executable, categories,
                     item_type, icon, icon_name, filename, False, show])

        if item_type == MenuItemTypes.DIRECTORY:  # type: ignore
            treestore = menu_to_treestore(treestore, treeiter, item[3])

    return treestore


def get_treestore():
    """Get the TreeStore implementation of the current menu."""
    treestore = Gtk.TreeStore(  # type: ignore
        str,  # Name
        str,  # Displayed Name
        str,  # Comment
        str,  # Executable
        str,  # Categories
        int,  # MenuItemType
        Gio.Icon,  # GIcon
        str,  # icon-name
        str,  # Filename
        bool,  # Expand
        bool  # Show
    )
    menus = get_menus()
    if not menus:
        return None
    return menu_to_treestore(treestore, None, menus[0])


def get_extended_icons(icon_names, extended_names):
    results = []
    prefer_symbolic = icon_theme_name in [
        "Adwaita"] and "folder" in extended_names
    for name in icon_names:
        if icon_theme.has_icon(name):
            return results
    if prefer_symbolic:
        for name in extended_names:
            name = name + "-symbolic"
            if icon_theme.has_icon(name):
                results.append(name)
        if len(results) > 0:
            return results
    for name in extended_names:
        if icon_theme.has_icon(name):
            results.append(name)
    return results


def get_submenus(menu, tree_dir):
    """Get the submenus for a tree directory."""
    structure = []
    for child in menu.getContents(tree_dir):
        if isinstance(child, GMenu.TreeSeparator):
            structure.append([MenuItemTypes.SEPARATOR, child, None, None])  # type: ignore
        else:
            if isinstance(child, GMenu.TreeEntry):
                item_type = MenuItemTypes.APPLICATION  # type: ignore
                entry_id = child.get_desktop_file_id()
                app_info = child.get_app_info()
                icon = app_info.get_icon()
                icon_name = app_info.get_string("Icon")
                extended_icon_names = [
                    "applications-other",
                    "application-x-executable"]
                display_name = app_info.get_display_name()
                generic_name = app_info.get_generic_name()
                comment = app_info.get_description()
                keywords = app_info.get_keywords()
                categories = app_info.get_categories()
                executable = app_info.get_executable()
                filename = child.get_desktop_file_path()
                submenus = None

                is_hidden = app_info.get_is_hidden()
                no_display = app_info.get_nodisplay()
                show_in = app_info.get_show_in()
                hidden = is_hidden or no_display or not show_in

            elif isinstance(child, GMenu.TreeDirectory):
                item_type = MenuItemTypes.DIRECTORY  # type: ignore
                entry_id = child.get_menu_id()
                icon = child.get_icon()
                icon_name = None
                extended_icon_names = ["folder"]
                display_name = child.get_name()
                generic_name = child.get_generic_name()
                comment = child.get_comment()
                keywords = []
                categories = ""
                executable = None
                filename = child.get_desktop_file_path()
                hidden = child.get_is_nodisplay()
                submenus = get_submenus(menu, child)

            else:
                continue

            icon_names = []
            if isinstance(icon, Gio.ThemedIcon):
                icon_names = icon.get_names()
            elif isinstance(icon, Gio.FileIcon):
                icon_names = [icon.get_file().get_path()]

            if icon_name is not None:
                icon_names = [icon_name] + icon_names

            for extended_icon_name in get_extended_icons(
                    icon_names, extended_icon_names):
                if isinstance(icon, Gio.ThemedIcon):
                    icon.append_name(extended_icon_name)
                icon_names.append(extended_icon_name)

            if icon_name is None:
                icon_name = icon_names[0]

            elif icon is None:
                icon = Gio.ThemedIcon.new(icon_name)

            if filename is not None:
                filename = os.path.realpath(filename)

            details = {'display_name': display_name,
                       'generic_name': generic_name,
                       'comment': comment,
                       'keywords': keywords,
                       'categories': categories,
                       'executable': executable,
                       'filename': filename,
                       'icon': icon,
                       'icon_name': icon_name,
                       'show': not hidden}
            entry = [item_type, entry_id, details, submenus]
            structure.append(entry)

    return structure


def get_menus():
    """Get the menus from the MenuEditor"""
    menu = MenuEditor()
    if not menu.loaded:
        return None
    structure = []
    toplevels = []
    global menu_name
    menu_name = menu.tree.get_root_directory().get_menu_id()
    for child in menu.getMenus(None):
        toplevels.append(child)
    for top in toplevels:
        structure.append(get_submenus(menu, top[0]))
    menu.unmap()
    return structure


def removeWhitespaceNodes(node):
    """Remove whitespace nodes from the xml dom."""
    remove_list = []
    for child in node.childNodes:
        if child.nodeType == xml.dom.minidom.Node.TEXT_NODE:
            child.data = child.data.strip()
            if not child.data.strip():
                remove_list.append(child)
        elif child.hasChildNodes():
            removeWhitespaceNodes(child)
    for node in remove_list:
        node.parentNode.removeChild(node)


def getUserMenuXml(tree):
    """Return the header portions of the menu xml file."""
    system_file = util.getSystemMenuPath(
        os.path.basename(tree.get_canonical_menu_path()))
    name = tree.get_root_directory().get_menu_id()
    menu_xml = "<!DOCTYPE Menu PUBLIC '-//freedesktop//DTD Menu 1.0//EN'" \
               " 'http://standards.freedesktop.org/menu-spec/menu-1.0.dtd'>\n"
    menu_xml += "<Menu>\n  <Name>" + name + "</Name>\n  "
    if system_file is not None:
        menu_xml += "<MergeFile type=\"parent\">" + system_file + "</MergeFile>\n"
    menu_xml += "</Menu>\n"
    return menu_xml


class MenuEditor(object):
    """MenuEditor class, adapted and minimized from Alacarte Menu Editor."""

    loaded = False

    def __init__(self, basename=None):
        """init"""

        # Remember to keep menulibre-menu-validate's GMenu object creation
        # in-sync with this code
        basename = basename or get_default_menu()
        if basename is None:
            return

        # For systems where desktop directories are installed in subdirectories,
        # we need to first bring them to the toplevel for GMenu to see them
        # correctly.
        mapDesktopEnvironmentDirectories()

        self.tree = GMenu.Tree.new(basename,
                                   GMenu.TreeFlags.SHOW_EMPTY |
                                   GMenu.TreeFlags.INCLUDE_EXCLUDED |
                                   GMenu.TreeFlags.INCLUDE_NODISPLAY |
                                   GMenu.TreeFlags.SHOW_ALL_SEPARATORS |
                                   GMenu.TreeFlags.SORT_DISPLAY_NAME)
        self.load()

        self.path = os.path.join(
            util.getUserMenusDirectory(), self.tree.props.menu_basename)
        logger.debug("Using menu: %s" % self.path)
        self.loadDOM()

        self.loaded = self.hasContents()

    def loadDOM(self):
        """loadDOM"""
        try:
            self.dom = xml.dom.minidom.parse(self.path)
        except (IOError, xml.parsers.expat.ExpatError):
            self.dom = xml.dom.minidom.parseString(
                getUserMenuXml(self.tree))
        removeWhitespaceNodes(self.dom)

    def load(self):
        """load"""
        if not self.tree.load_sync():
            raise ValueError("can not load menu tree %r" %
                             (self.tree.props.menu_basename,))

    def unmap(self):
        unmapDesktopEnvironmentDirectories()

    def hasContents(self):
        for child in self.getMenus(None):
            submenus = self.getContents(child[0])
            if len(submenus) > 0:
                return True
        return False

    def getMenus(self, parent):
        """getMenus"""
        if parent is None:
            yield (self.tree.get_root_directory(), True)
            return

        item_iter = parent.iter()
        item_type = item_iter.next()
        while item_type != GMenu.TreeItemType.INVALID:
            if item_type == GMenu.TreeItemType.DIRECTORY:
                item = item_iter.get_directory()
                yield (item, True)
            item_type = item_iter.next()

    def getContents(self, item):
        """getContents"""
        contents = []
        item_iter = item.iter()
        item_type = item_iter.next()

        found_directories = []

        while item_type != GMenu.TreeItemType.INVALID:
            item = None
            if item_type == GMenu.TreeItemType.DIRECTORY:
                item = item_iter.get_directory()
                desktop = item.get_desktop_file_path()
                if desktop is not None:
                    desktop = os.path.realpath(desktop)
                if desktop is None or desktop in found_directories:
                    # Do not include directories without filenames.
                    # Do not include duplicate directories.
                    item = None
                else:
                    found_directories.append(desktop)
            elif item_type == GMenu.TreeItemType.ENTRY:
                item = item_iter.get_entry()
            elif item_type == GMenu.TreeItemType.HEADER:
                item = item_iter.get_header()
            elif item_type == GMenu.TreeItemType.ALIAS:
                item = item_iter.get_alias()
            elif item_type == GMenu.TreeItemType.SEPARATOR:
                item = item_iter.get_separator()
            if item:
                contents.append(item)
            item_type = item_iter.next()
        return contents



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk  # type: ignore

class LabelWithHidingButton(Gtk.EventBox):

    def __init__(self, label, icon_name, icon_size):
        super().__init__()

        box = Gtk.Box.new(orientation=Gtk.Orientation.HORIZONTAL, spacing=3)
        self.add(box)

        self._label = Gtk.Label.new(label)
        self._label.set_xalign(0.0)

        box.add(self._label)

        self._button = Gtk.Button.new_from_icon_name(icon_name, icon_size)
        self._button.set_name('hideybutton')
        self._button.set_opacity(0.0)

        context = self._button.get_style_context()
        context.add_class("flat")

        box.add(self._button)

        self.add_events(Gdk.EventMask.ENTER_NOTIFY_MASK |
                        Gdk.EventMask.LEAVE_NOTIFY_MASK)

        self._label_entered = False
        self._button_entered = False
        self._button_focused = False

        self.connect("enter-notify-event", self._enter_notify_cb, self._button)
        self.connect("leave-notify-event", self._leave_notify_cb, self._button)

        self._button.connect(
            "enter-notify-event",
            self._button_enter_notify_cb,
            self._button)
        self._button.connect(
            "leave-notify-event",
            self._button_leave_notify_cb,
            self._button)

        self._button.connect(
            "focus-in-event",
            self._button_focus_in_cb,
            self._button)
        self._button.connect(
            "focus-out-event",
            self._button_focus_out_cb,
            self._button)

    def get_button(self):
        return self._button

    def get_label(self):
        return self._label

    def _toggle_button_visibility(self, button):
        if self._label_entered or self._button_entered or self._button_focused:
            button.set_opacity(1.0)
        else:
            button.set_opacity(0.0)

    def _enter_notify_cb(self, widget, event, button):
        self._label_entered = True
        self._toggle_button_visibility(button)

    def _leave_notify_cb(self, widget, event, button):
        self._label_entered = False
        self._toggle_button_visibility(button)

    def _button_enter_notify_cb(self, widget, event, button):
        self._button_entered = True
        self._toggle_button_visibility(button)

    def _button_leave_notify_cb(self, widget, event, button):
        self._button_entered = False
        self._toggle_button_visibility(button)

    def _button_focus_in_cb(self, widget, event, button):
        self._button_focused = True
        self._toggle_button_visibility(button)

    def _button_focus_out_cb(self, widget, event, button):
        self._button_focused = False
        self._toggle_button_visibility(button)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#   Copyright (C) 2017-2018 OmegaPhil <OmegaPhil@startmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
import subprocess

import getpass
import psutil
import shutil

from locale import gettext as _

import gi
gi.require_version('Gdk', '3.0')
from gi.repository import GLib, Gdk  # type: ignore

logger = logging.getLogger('menulibre')

old_psutil_format = isinstance(psutil.Process.username, property)


def enum(**enums):
    """Add enumerations to Python."""
    return type('Enum', (), enums)


MenuItemTypes = enum(
    SEPARATOR=-1,
    APPLICATION=0,
    LINK=1,
    DIRECTORY=2
)


MenuItemKeys = (
    # Key, Type, Required, Types (MenuItemType)
    ("Version", str, False, (0, 1, 2)),
    ("Type", str, True, (0, 1, 2)),
    ("Name", str, True, (0, 1, 2)),
    ("GenericName", str, False, (0, 1, 2)),
    ("NoDisplay", bool, False, (0, 1, 2)),
    ("Comment", str, False, (0, 1, 2)),
    ("Icon", str, False, (0, 1, 2)),
    ("Hidden", bool, False, (0, 1, 2)),
    ("OnlyShowIn", list, False, (0, 1, 2)),
    ("NotShowIn", list, False, (0, 1, 2)),
    ("DBusActivatable", bool, False, (0,)),
    ("PrefersNonDefaultGPU", bool, False, (0,)),
    ("X-GNOME-UsesNotifications", bool, False, (0,)),
    ("TryExec", str, False, (0,)),
    ("Exec", str, True, (0,)),
    ("Path", str, False, (0,)),
    ("Terminal", bool, False, (0,)),
    ("Actions", str, False, (0,)),
    ("MimeType", list, False, (0,)),
    ("Categories", list, False, (0,)),
    ("Implements", list, False, (0,)),
    ("Keywords", list, False, (0,)),
    ("StartupNotify", bool, False, (0,)),
    ("StartupWMClass", str, False, (0,)),
    ("URL", str, True, (1,))
)


def getRelatedKeys(menu_item_type, key_only=False):
    if isinstance(menu_item_type, str):
        if menu_item_type == "Application":
            menu_item_type = MenuItemTypes.APPLICATION  # type: ignore
        elif menu_item_type == "Link":
            menu_item_type = MenuItemTypes.LINK  # type: ignore
        elif menu_item_type == "Directory":
            menu_item_type = MenuItemTypes.DIRECTORY  # type: ignore

    results = []
    for tup in MenuItemKeys:
        if menu_item_type in tup[3]:
            if key_only:
                results.append(tup[0])
            else:
                results.append((tup[0], tup[1], tup[2]))
    return results


def escapeText(text):
    if text is None:
        return ""
    return GLib.markup_escape_text(text)


def getProcessUsername(process):
    """Get the username of the process owner. Return None if fail."""
    username = None

    try:
        if old_psutil_format:
            username = process.username
        else:
            username = process.username()
    except:  # noqa
        pass

    return username


def getProcessName(process):
    """Get the process name. Return None if fail."""
    p_name = None

    try:
        if old_psutil_format:
            p_name = process.name
        else:
            p_name = process.name()
    except:  # noqa
        pass

    return p_name


def getProcessList():
    """Return a list of unique process names for the current user."""
    username = getpass.getuser()
    try:
        pids = psutil.get_pid_list()  # type: ignore
    except AttributeError:
        pids = psutil.pids()
    processes = []
    for pid in pids:
        try:
            process = psutil.Process(pid)
            p_user = getProcessUsername(process)
            if p_user == username:
                p_name = getProcessName(process)
                if p_name is not None and p_name not in processes:
                    processes.append(p_name)
        except:  # noqa
            pass
    processes.sort()
    return processes


def getRelativeName(filename: str):
    if filename.endswith('.desktop'):
        basename = filename.split('/applications/', 1)[1]
    elif filename.endswith('.directory'):
        basename = filename.split('/desktop-directories/', 1)[1]
    else:
        basename = ""
    return basename


def getBasename(filename: str):
    if filename.endswith('.desktop'):
        basename = filename.split('/applications/', 1)[1]
    elif filename.endswith('.directory'):
        basename = filename.split('/desktop-directories/', 1)[1]
        if basename.startswith("%s/" % getDefaultMenuName()):
            basename = filename.split("%s/" % getDefaultMenuName(), 1)[1]
    else:
        basename = ""
    return basename


def getCurrentDesktop():
    current_desktop = os.environ.get("XDG_CURRENT_DESKTOP", "")
    current_desktop = current_desktop.lower()
    for desktop in ["budgie", "pantheon", "gnome"]:
        if desktop in current_desktop:
            return desktop
    if "kde" in current_desktop:
        kde_version = int(os.environ.get("KDE_SESSION_VERSION", "4"))
        if kde_version >= 5:
            return "plasma"
        return "kde"
    return current_desktop


def getDefaultMenuName():
    prefix = getDefaultMenuPrefix()
    if prefix.endswith("-"):
        return prefix[:-1]
    return prefix


def getDefaultMenuPrefix():  # noqa
    """Return the default menu prefix."""
    prefix = os.environ.get('XDG_MENU_PREFIX', '')

    # Cinnamon and MATE don't set this variable
    if prefix == "":
        if 'cinnamon' in os.environ.get('DESKTOP_SESSION', ''):
            prefix = 'cinnamon-'
        elif 'mate' in os.environ.get('DESKTOP_SESSION', ''):
            prefix = 'mate-'
        # Somehow the XDG_MENU_PREFIX isn't exposed in Ubuntu Unity
        elif 'unity' in os.environ.get('DESKTOP_SESSION', ''):
            prefix = 'gnome-'

    if prefix == "":
        desktop = getCurrentDesktop()
        if desktop == 'plasma':
            prefix = 'kf5-'
        elif desktop == 'kde':
            prefix = 'kde4-'

    if prefix == "":
        processes = getProcessList()
        if 'xfce4-panel' in processes:
            prefix = 'xfce-'
        elif 'mate-panel' in processes:
            prefix = 'mate-'

    if len(prefix) == 0:
        logger.warning("No menu prefix found, MenuLibre will not function "
                       "properly.")

    return prefix


def getMenuDiagnostics():
    diagnostics = {}
    keys = [
        "XDG_CURRENT_DESKTOP",
        "XDG_MENU_PREFIX",
        "DESKTOP_SESSION",
        "KDE_SESSION_VERSION"
    ]
    for k in keys:
        diagnostics[k] = os.environ.get(k, "None")

    menu_dirs = [
        getUserMenusDirectory()
    ]
    for path in GLib.get_system_config_dirs():
        menu_dirs.append(os.path.join(path, 'menus'))
    menus = []
    for menu_dir in menu_dirs:
        try:
            for filename in os.listdir(menu_dir):
                if filename.endswith(".menu"):
                    menus.append(os.path.join(menu_dir, filename))
        except FileNotFoundError:
            pass
    menus.sort()

    diagnostics["MENUS"] = ", ".join(menus)

    return diagnostics


def getItemSearchPaths():
    search_paths = []
    for path in GLib.get_system_data_dirs():
        search_paths.append(os.path.join(path, 'applications'))
    return search_paths


def getItemPath(file_id):
    """Return the path to the system-installed .desktop file."""
    for path in getItemSearchPaths():
        file_path = os.path.join(path, file_id)
        if os.path.isfile(file_path):
            return file_path
    return None


def getUserApplicationsDirectory():
    """Return the path to the user applications directory."""
    item_dir = os.path.join(GLib.get_user_data_dir(), 'applications')
    if not os.path.isdir(item_dir):
        os.makedirs(item_dir)
    return item_dir


def getUserItemPath(file_id):
    """Return the path to the system-installed .desktop file."""
    path = getUserApplicationsDirectory()
    file_path = os.path.join(path, file_id)
    if os.path.isfile(file_path):
        return file_path
    return None


def getDirectorySearchPaths():
    search_paths = []
    for path in GLib.get_system_data_dirs():
        search_paths.append(os.path.join(path, 'desktop-directories'))
    return search_paths


def getDirectoryPath(file_id):
    """Return the path to the system-installed .directory file."""
    for path in getDirectorySearchPaths():
        file_path = os.path.join(path, file_id)
        if os.path.isfile(file_path):
            return file_path
    return None


def mapDesktopEnvironmentDirectories():
    """
    This feels wrong, but to make GMenu correctly handle desktop directories
    in subdirectories, we need to bring them up to the top level.
    """
    menu = getDefaultMenuName()
    if len(menu) == 0:
        return

    file_ids = []
    for path in getDirectorySearchPaths():
        if not os.path.exists(path):
            continue
        for filename in os.listdir(path):
            if filename.endswith(".desktop"):
                file_ids.append(filename)

    de_paths = {}
    for path in GLib.get_system_data_dirs():
        de_path = os.path.join(path, menu, 'desktop-directories')
        if not os.path.exists(de_path):
            continue
        for filename in os.listdir(de_path):
            if filename not in file_ids:
                de_paths[filename] = de_path

    user_dir = getUserDirectoriesDirectory()
    for filename, basedir in de_paths.items():
        if filename in file_ids:
            continue

        target_dir = os.path.join(user_dir, menu)
        try:
            os.makedirs(target_dir)
        except BaseException:
            pass
        target = os.path.join(target_dir, filename)
        src = os.path.join(basedir, filename)
        if not os.path.exists(target):
            try:
                logger.debug("copy %s to %s" % (src, target))
                shutil.copy2(src, target)
            except BaseException:
                logger.warning("Failed to copy %s to %s" % (src, target))
                continue

        try:
            symlink = os.path.join(user_dir, filename)
            if os.path.exists(symlink):
                continue
            os.symlink(target, symlink)
        except BaseException:
            logger.warning("Failed to symlink %s to %s" % (src, target))


def unmapDesktopEnvironmentDirectories():
    """
    This feels wrong, but to make GMenu correctly handle desktop directories
    in subdirectories, we need to bring them up to the top level.
    """
    menu = getDefaultMenuName()
    if len(menu) == 0:
        return

    user_dir = getUserDirectoriesDirectory()

    de_dir = os.path.join(user_dir, menu)
    if not os.path.exists(de_dir):
        return

    for filename in os.listdir(user_dir):
        filename = os.path.join(user_dir, filename)

        if not filename.endswith(".directory"):
            continue
        if not os.path.islink(filename):
            continue

        realpath = os.path.realpath(filename)
        if not realpath.startswith(de_dir):
            continue

        try:
            os.remove(filename)
        except BaseException:
            logger.warning("Failed to remove symlink %s" % filename)


def getUserDirectoriesDirectory():
    """Return the path to the user desktop-directories directory."""
    menu_dir = os.path.join(GLib.get_user_data_dir(), 'desktop-directories')
    if not os.path.isdir(menu_dir):
        os.makedirs(menu_dir)
    return menu_dir


def getUserDirectoryPath(file_id):
    """Return the path to the system-installed .directory file."""
    path = getUserDirectoriesDirectory()
    file_path = os.path.join(path, file_id)
    if os.path.isfile(file_path):
        return file_path
    return None


def getUserMenusDirectory():
    """Return the path to the user menus directory."""
    menu_dir = os.path.join(GLib.get_user_config_dir(), 'menus')
    if not os.path.isdir(menu_dir):
        os.makedirs(menu_dir)
    return menu_dir


def getUserLauncherPath(basename):
    """Return the user-installed path to a .desktop or .directory file."""
    if basename.endswith('.desktop'):
        return getUserItemPath(basename)
    else:
        return getUserDirectoryPath(basename)


def getSystemMenuPath(file_id):
    """Return the path to the system-installed menu file."""
    for path in GLib.get_system_config_dirs():
        file_path = os.path.join(path, 'menus', file_id)
        if os.path.isfile(file_path):
            return file_path
    return None


def getSystemLauncherPath(basename):
    """Return the system-installed path to a .desktop or .directory file."""
    if basename.endswith('.desktop'):
        return getItemPath(basename)
    else:
        return getDirectoryPath(basename)


def getDirectoryName(directory_str):  # noqa
    """Return the directory name to be used in the XML file."""

    # Note: When adding new logic here, please see if
    # getDirectoryNameFromCategory should also be updated

    # Get the menu prefix
    prefix = getDefaultMenuPrefix()
    has_prefix = False

    basename = getBasename(directory_str)
    name, ext = os.path.splitext(basename)

    # Handle directories like xfce-development
    if name.startswith(prefix):
        name = name[len(prefix):]
        name = name.title()
        has_prefix = True

    # Handle X-GNOME, X-XFCE
    if name.startswith("X-"):
        # Handle X-GNOME, X-XFCE
        condensed = name.split('-', 2)[-1]
        non_camel = re.sub('(?!^)([A-Z]+)', r' \1', condensed)
        return non_camel

    # Cleanup ArcadeGames and others as per the norm.
    if name.endswith('Games') and name != 'Games':
        condensed = name[:-5]
        non_camel = re.sub('(?!^)([A-Z]+)', r' \1', condensed)
        return non_camel

    # GNOME...
    if name == 'AudioVideo' or name == 'Audio-Video':
        return 'Multimedia'

    if name == 'Game':
        return 'Games'

    if name == 'Network' and prefix != 'xfce-':
        return 'Internet'

    if name == 'Utility':
        return 'Accessories'

    if name == 'System-Tools':
        if prefix == 'lxde-':
            return 'Administration'
        else:
            return 'System'

    if name == 'Settings':
        if prefix == 'lxde-':
            return 'DesktopSettings'
        elif has_prefix and prefix == 'xfce-':
            return name
        else:
            return 'Preferences'

    if name == 'Settings-System':
        return 'Administration'

    if name == 'GnomeScience':
        return 'Science'

    if name == 'Utility-Accessibility':
        return 'Universal Access'

    # We tried, just return the name.
    return name


def getDirectoryNameFromCategory(name):  # noqa
    """Guess at the directory name a category should cause its launcher to
    appear in. This is used to add launchers to or remove from the right
    directories after category addition without having to restart menulibre."""

    # Note: When adding new logic here, please see if
    # getDirectoryName should also be updated

    # I don't want to overload the use of getDirectoryName, so have spun out
    # this similar function

    # Only interested in generic categories here, so no need to handle
    # categories named after desktop environments
    prefix = getDefaultMenuPrefix()

    # Cleanup ArcadeGames and others as per the norm.
    if name.endswith('Games') and name != 'Games':
        condensed = name[:-5]
        non_camel = re.sub('(?!^)([A-Z]+)', r' \1', condensed)
        return non_camel

    # GNOME...
    if name == 'AudioVideo' or name == 'Audio-Video':
        return 'Multimedia'

    if name == 'Game':
        return 'Games'

    if name == 'Network':
        return 'Internet'

    if name == 'Utility':
        return 'Accessories'

    if name == 'System-Tools':
        if prefix == 'lxde-':
            return 'Administration'
        else:
            return 'System'

    if name == 'Settings':
        if prefix == 'lxde-':
            return 'DesktopSettings'
        elif prefix == 'xfce-':
            return name
        else:
            return 'Preferences'

    if name == 'Settings-System':
        return 'Administration'

    if name == 'GnomeScience':
        return 'Science'

    if name == 'Utility-Accessibility':
        return 'Universal Access'

    # We tried, just return the name.
    return name


def getRequiredCategories(directory):
    """Return the list of required categories for a directory string."""
    prefix = getDefaultMenuPrefix()
    if directory is not None:
        basename = getBasename(directory)
        name, ext = os.path.splitext(basename)

        # Handle directories like xfce-development
        if name.startswith(prefix):
            name = name[len(prefix):]
            name = name.title()

        if name == 'Accessories':
            return ['Utility']

        if name == 'Games':
            return ['Game']

        if name == 'Multimedia':
            return ['AudioVideo']

        else:
            return [name]
    else:
        # Get The Toplevel item if necessary...
        if prefix == 'xfce-':
            return ['X-XFCE', 'X-Xfce-Toplevel']
    return []


def getSaveFilename(name, filename, item_type, force_update=False):  # noqa
    """Determime the filename to be used to store the launcher.

    Return the filename to be used."""
    # Check if the filename is writeable. If not, generate a new one.
    unique = filename is None or len(filename) == 0

    if unique or not os.access(filename, os.W_OK):
        # No filename, make one from the launcher name.
        if unique:
            basename = "menulibre-" + name.lower().replace(' ', '-')

        # Use the current filename as a base.
        else:
            basename = getRelativeName(filename)

        # Split the basename into filename and extension.
        name, ext = os.path.splitext(basename)

        # Get the save location of the launcher base on type.
        if item_type == 'Application':
            path = getUserApplicationsDirectory()
            ext = '.desktop'
        elif item_type == 'Directory':
            path = getUserDirectoriesDirectory()
            ext = '.directory'
        else:
            path = ""

        basedir = os.path.dirname(os.path.join(path, basename))
        if not os.path.exists(basedir):
            os.makedirs(basedir)

        # Index for unique filenames.
        count = 1

        # Be sure to not overwrite system launchers if new.
        if unique:
            # Check for the system version of the launcher.
            if getSystemLauncherPath("%s%s" % (name, ext)) is not None:
                # If found, check for any additional ones.
                while getSystemLauncherPath("%s%i%s" % (name, count, ext)) \
                        is not None:
                    count += 1

                # Now be sure to not overwrite locally installed ones.
                filename = os.path.join(path, name)
                filename = "%s%i%s" % (filename, count, ext)

                # Append numbers as necessary to make the filename unique.
                while os.path.exists(filename):
                    new_basename = "%s%i%s" % (name, count, ext)
                    filename = os.path.join(path, new_basename)
                    count += 1

            else:
                # Create the new base filename.
                filename = os.path.join(path, name)
                filename = "%s%s" % (filename, ext)

                # Append numbers as necessary to make the filename unique.
                while os.path.exists(filename):
                    new_basename = "%s%i%s" % (name, count, ext)
                    filename = os.path.join(path, new_basename)
                    count += 1

        else:
            # Create the new base filename.
            filename = os.path.join(path, basename)

            if force_update:
                return filename

            # Append numbers as necessary to make the filename unique.
            while os.path.exists(filename):
                new_basename = "%s%i%s" % (name, count, ext)
                filename = os.path.join(path, new_basename)
                count += 1

    return filename


def check_keypress(event, keys):  # noqa
    """Compare keypress events with desired keys and return True if matched."""
    if 'Control' in keys:
        if not bool(event.get_state() & Gdk.ModifierType.CONTROL_MASK):
            return False
    if 'Alt' in keys:
        if not bool(event.get_state() & Gdk.ModifierType.MOD1_MASK):
            return False
    if 'Shift' in keys:
        if not bool(event.get_state() & Gdk.ModifierType.SHIFT_MASK):
            return False
    if 'Super' in keys:
        if not bool(event.get_state() & Gdk.ModifierType.SUPER_MASK):
            return False
    if 'Escape' in keys:
        keys[keys.index('Escape')] = 'escape'
    keyval_name = Gdk.keyval_name(event.get_keyval()[1])
    if keyval_name is None:
        return False
    if keyval_name.lower() not in keys:
        return False

    return True


def determine_bad_desktop_files():
    """Run the gmenu-invalid-desktop-files script to get at the GMenu library's
    debug output, which lists files that failed to load, and return these as a
    sorted list."""

    # Run the helper script with normal binary lookup via the shell, capturing
    # stderr, sensitive to errors
    try:
        result = subprocess.run(['menulibre-menu-validate'],
                                stderr=subprocess.PIPE, shell=True, check=True)
    except subprocess.CalledProcessError:
        return []

    # stderr is returned as bytes, so converting it to the line-buffered output
    # I actually want
    bad_desktop_files = []
    for line in result.stderr.decode('UTF-8').split('\n'):
        matches = re.match(r'^Failed to load "(.+\.desktop)"$', line)
        if matches:
            desktop_file = matches.groups()[0]
            if validate_desktop_file(desktop_file):
                bad_desktop_files.append(matches.groups()[0])

    # Alphabetical sort on bad desktop file paths
    bad_desktop_files.sort()

    return bad_desktop_files


def find_program(program):
    program = program.strip()
    if len(program) == 0:
        return None

    params = list(GLib.shell_parse_argv(program)[1])  # type: ignore
    executable = params[0]

    if os.path.exists(executable):
        return executable

    path = GLib.find_program_in_path(executable)
    if path is not None:
        return path

    return None


def validate_desktop_file(desktop_file):  # noqa
    """Validate a known-bad desktop file in the same way GMenu/glib does, to
    give a user real information about why certain files are broken."""

    # This is a reimplementation of the validation logic in glib2's
    # gio/gdesktopappinfo.c:g_desktop_app_info_load_from_keyfile.
    # gnome-menus appears also to try to do its own validation in
    # libmenu/desktop-entries.c:desktop_entry_load, however
    # g_desktop_app_info_new_from_filename will not return a valid
    # GDesktopAppInfo in the first place if something is wrong with the
    # desktop file

    try:

        # Looks like load_from_file is not a class method??
        keyfile = GLib.KeyFile()
        keyfile.load_from_file(desktop_file, GLib.KeyFileFlags.NONE)

    except Exception as e:
        # Translators: This error is displayed when a desktop file cannot
        # be correctly read by MenuLibre. A (possibly untranslated) error
        # code is displayed.
        return _('Unable to load desktop file due to the following error:'
                 ' %s') % e

    # File is at least a valid keyfile, so can start the real desktop
    # validation
    # Start group validation
    try:
        start_group = keyfile.get_start_group()
    except GLib.Error:
        start_group = None

    if start_group != GLib.KEY_FILE_DESKTOP_GROUP:
        # Translators: This error is displayed when the first group in a
        # failing desktop file is incorrect. "Start group" can be safely
        # translated.
        return (_('Start group is invalid - currently \'%s\', should be '
                  '\'%s\'') % (start_group, GLib.KEY_FILE_DESKTOP_GROUP))

    # Type validation
    try:
        type_key = keyfile.get_string(start_group,  # type: ignore
                                      GLib.KEY_FILE_DESKTOP_KEY_TYPE)
    except GLib.Error:
        # Translators: This error is displayed when a required key is
        # missing in a failing desktop file.
        return _('%s key not found') % 'Type'

    valid_type_keys = [
        GLib.KEY_FILE_DESKTOP_TYPE_APPLICATION,
        GLib.KEY_FILE_DESKTOP_TYPE_LINK,
        GLib.KEY_FILE_DESKTOP_TYPE_DIRECTORY,
        # KDE-specific types
        # https://specifications.freedesktop.org/desktop-entry-spec/latest/apb.html
        "ServiceType", "Service", "FSDevice"
    ]
    if type_key not in valid_type_keys:
        # Translators: This error is displayed when a failing desktop file
        # has an invalid value for the provided key.
        return (_('%s value is invalid - currently \'%s\', should be \'%s\'')
                % ('Type', type_key, GLib.KEY_FILE_DESKTOP_TYPE_APPLICATION))

    # Validating 'try exec' if its present
    # Invalid TryExec is a valid state. "If the file is not present or if it
    # is not executable, the entry may be ignored (not be used in menus, for
    # example)."
    try:
        try_exec = keyfile.get_string(start_group,  # type: ignore
                                      GLib.KEY_FILE_DESKTOP_KEY_TRY_EXEC)
    except GLib.Error:
        pass

    else:
        try:
            if len(try_exec) > 0 and find_program(try_exec) is None:
                return False
        except Exception as e:
            return False

    # Validating executable
    try:
        exec_key = keyfile.get_string(start_group,  # type: ignore
                                      GLib.KEY_FILE_DESKTOP_KEY_EXEC)
    except GLib.Error:
        return False  # LP: #1788814, Exec key is not required

    try:
        if find_program(exec_key) is None:
            return (_('%s program \'%s\' has not been found in the PATH')
                    % ('Exec', exec_key))
    except Exception as e:
        return (_('%s program \'%s\' is not a valid shell command '
                  'according to GLib.shell_parse_argv, error: %s')
                % ('Exec', exec_key, e))

    if type_key == "Service":
        return False  # KDE services are not displayed in the menu

    # Translators: This error is displayed for a failing desktop file where
    # errors were detected but the file seems otherwise valid.
    return _('Unknown error. Desktop file appears to be valid.')



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#   Copyright (C) 2016 OmegaPhil <OmegaPhil@startmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

# lint:disable
try:
    import xml.etree.cElementTree
    from xml.etree.cElementTree import ElementTree, Element, SubElement
except ImportError:
    import xml.etree.ElementTree  # noqa
    from xml.etree.ElementTree import ElementTree, Element, SubElement
# lint:enable

import os

from . import util
from .util import MenuItemTypes

from . import MenuEditor

# Store user desktop directory location
directories = util.getUserDirectoriesDirectory()


# Prevent gnome-menus crash
processed_directories = []


def indent(elem, level=0):
    """Indentation code to make XML output easier to read."""
    i = "\n" + level * "\t"
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "\t"
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


class XmlMenuElement(Element):
    """An extension of the ElementTree.Element class which adds Menu
    constructing functionality."""

    def __init__(self, *args, **kwargs):
        """Initialize the XmlMenuElement class.  This takes all regular
        arguments for ElementTree.Element, as well as the menu_name keyword
        argument, used for setting the Menu Name property."""
        if 'menu_name' in list(kwargs.keys()):
            menu_name = kwargs['menu_name']
            kwargs.pop("menu_name", None)
        else:
            menu_name = None
        super(XmlMenuElement, self).__init__(*args, **kwargs)
        if menu_name:
            SubElement(self, "Name").text = menu_name

    def addMenu(self, menu_name, filename=None):
        """Add a submenu <Menu> to this XmlMenuElement.

        Return a reference to that submenu XmlMenuElement."""
        menu = XmlMenuElement("Menu")
        self.append(menu)
        SubElement(menu, "Name").text = menu_name
        if filename:
            SubElement(menu, "Directory").text = os.path.basename(filename)
            realpath = os.path.realpath(filename)
            realdir = os.path.dirname(realpath)
            if realdir.startswith(directories):
                SubElement(menu, "DirectoryDir").text = realdir
            elif filename.startswith(directories):
                SubElement(menu, "DirectoryDir").text = directories
        return menu

    def addMenuname(self, menu_name):
        """Add a menuname <Menuname> to this XmlMenuElement.

        Return a reference to that menuname XmlMenuElement."""
        element = XmlMenuElement("Menuname")
        element.text = menu_name
        self.append(element)
        return element

    def addSeparator(self):
        """Add a <Separator> element to this XmlMenuElement.

        Return a reference to that separator XmlMenuElement."""
        element = XmlMenuElement("Separator")
        self.append(element)
        return element

    def addMergeFile(self, filename, merge_type="parent"):
        """Add a merge file <MergeFile> to this XmlMenuElement.

        Return a reference to that merge file XmlMenuElement."""
        element = XmlMenuElement("MergeFile", type=merge_type)
        element.text = filename
        self.append(element)
        return element

    def addMerge(self, merge_type):
        """Add a merge <Merge> to this XmlMenuElement.

        Return a reference to that merge SubElement."""
        return SubElement(self, "Merge", type=merge_type)

    def addLayout(self):
        """Add a layout <Layout> to this XmlMenuElement.

        Return a reference to that layout XmlMenuElement"""
        element = XmlMenuElement("Layout")
        self.append(element)
        return element

    def addFilename(self, filename):
        """Add a filename <Filename> to this XmlMenuElement.

        Return a reference to that filename XmlMenuElement."""
        element = XmlMenuElement("Filename")
        element.text = filename
        self.append(element)
        return element

    def addInclude(self):
        """Add an include <Include> to this XmlMenuElement.

        Return a reference to that include XmlMenuElement"""
        element = XmlMenuElement("Include")
        self.append(element)
        return element

    def addCategory(self, category):
        """Add a category <Category> to this XmlMenuElement. Used primarily for
        Include and Exclude.

        Return a reference to that category XmlMenuElement."""
        element = XmlMenuElement("Category")
        element.text = category
        self.append(element)
        return element

    def addDefaults(self):
        """Add Default Menu Items"""
        SubElement(self, "DefaultAppDirs")
        SubElement(self, "DefaultDirectoryDirs")
        SubElement(self, "DefaultMergeDirs")


class XmlMenuElementTree(ElementTree):
    """An extension of the ElementTree.ElementTree class which simplifies the
    creation of FD.o Menus."""

    def __init__(self, menu_name, merge_file=None):
        """Initialize the XmlMenuElementTree class.  Accepts two arguments:
        - menu_name    Name of the menu (e.g. Xfce, Gnome)
        - merge_file   Merge file, used for extending an existing menu."""
        root = XmlMenuElement("Menu", menu_name=menu_name)
        root.addDefaults()

        # Xfce toplevel support
        if menu_name == 'Xfce':
            include = root.addInclude()
            include.addCategory("X-Xfce-Toplevel")

        if merge_file:
            root.addMergeFile(merge_file)
        super().__init__(root)

    def write(self, output_file):
        """Override for the ElementTree.write function. This variation adds
        the menu specification headers and writes the output in an
        easier-to-read format."""
        header = """<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE Menu
  PUBLIC '-//freedesktop//DTD Menu 1.0//EN'
  'http://standards.freedesktop.org/menu-spec/menu-1.0.dtd'>
"""
        copy = self.getroot()
        indent(copy)
        with open(output_file, 'w') as f:
            f.write(header)
            ElementTree(copy).write(f, encoding='unicode')


def model_to_xml_menus(model, model_parent=None, menu_parent=None):
    """Append the <Menu> elements to menu_parent."""
    for n_child in range(model.iter_n_children(model_parent)):
        treeiter = model.iter_nth_child(model_parent, n_child)

        # Extract the menu item details.
        name = model[treeiter][MenuEditor.COL_NAME]
        item_type = model[treeiter][MenuEditor.COL_TYPE]
        desktop = model[treeiter][MenuEditor.COL_FILENAME]

        if item_type == MenuItemTypes.DIRECTORY:  # type: ignore
            # Do not save duplicate directories.
            global processed_directories
            if desktop in processed_directories:
                continue

            # Add a menu child.
            if desktop is None:
                # Cinnamon fix.
                if name == 'wine-wine':
                    next_element = menu_parent.addMenu(name)  # type: ignore
                else:
                    continue
            else:
                directory_name = util.getDirectoryName(desktop)
                next_element = menu_parent.addMenu(directory_name, desktop)  # type: ignore

            # Do Menus
            model_to_xml_menus(model, treeiter, next_element)

            # Do Includes to allow for alacarte-created entries without
            # categories to persist (see LP: #1315880)
            model_to_xml_includes(model, treeiter, next_element)

            # Do Layouts
            model_to_xml_layout(model, treeiter, next_element)

        elif item_type == MenuItemTypes.APPLICATION:  # type: ignore
            pass

        elif item_type == MenuItemTypes.SEPARATOR:  # type: ignore
            pass


def model_to_xml_includes(model, model_parent=None, menu_parent=None):
    """Append <Include> elements for any application items that lack categories
    in a system directory (e.g. alacarte-created entries), and all items in
    custom directories."""

    # Looping for all items in directory
    for n_child in range(model.iter_n_children(model_parent)):
        treeiter = model.iter_nth_child(model_parent, n_child)

        # Extract the menu item details.
        categories = model[treeiter][MenuEditor.COL_CATEGORIES]
        item_type = model[treeiter][MenuEditor.COL_TYPE]
        desktop = model[treeiter][MenuEditor.COL_FILENAME]

        if desktop is None:
            continue

        # Detecting custom user directories
        user_directory = False
        if model_parent and categories:
            for category in categories.split(';'):
                if category.startswith('menulibre-'):
                    user_directory = True
                    break

        # Items in custom directories by menulibre have a category, but
        # includes are required otherwise they are dropped by GMenu
        if item_type == MenuItemTypes.APPLICATION and (  # type: ignore
                not categories or user_directory):
            include = menu_parent.addInclude()  # type: ignore
            try:
                include.addFilename(os.path.basename(desktop))
            except AttributeError:
                pass


def model_to_xml_layout(model, model_parent=None, menu_parent=None,  # noqa
                        merge=True):
    """Append the <Layout> element to menu_parent."""
    layout = menu_parent.addLayout()  # type: ignore

    # Add a merge for any submenus (except toplevel)
    if merge:
        layout.addMerge("menus")

    for n_child in range(model.iter_n_children(model_parent)):
        treeiter = model.iter_nth_child(model_parent, n_child)

        # Extract the menu item details.
        name = model[treeiter][MenuEditor.COL_NAME]
        item_type = model[treeiter][MenuEditor.COL_TYPE]
        desktop = model[treeiter][MenuEditor.COL_FILENAME]

        if item_type == MenuItemTypes.DIRECTORY:  # type: ignore
            # Do not save duplicate directories.
            global processed_directories
            if desktop in processed_directories:
                continue
            else:
                processed_directories.append(desktop)

            if desktop is None:
                # Cinnamon fix.
                if name == 'wine-wine':
                    layout.addMenuname(name)
                else:
                    continue
            else:
                directory_name = util.getDirectoryName(desktop)
                layout.addMenuname(directory_name)

        elif item_type == MenuItemTypes.APPLICATION and desktop is not None:  # type: ignore
            try:

                # According to the spec, desktop files may be located in
                # subdirectories of the '*/applications' directory that
                # effectively gives the contained desktop filenames an extra
                # prefix based on that directory name (this affects me with the
                # kde4 subdirectory desktop files). This only seems to matter
                # for layout generation - if you don't specify the desktop file
                # with the prefix here, the prefixed desktop file will not
                # match in the layout node when the menu is being constructed
                # See LP: #1315536 comment 5
                containing_dir = os.path.basename(os.path.dirname(desktop))
                desktop_filename = os.path.basename(desktop)
                if containing_dir != 'applications':
                    desktop_filename = '%s-%s' % (containing_dir,
                                                  desktop_filename)
                layout.addFilename(desktop_filename)
            except AttributeError:
                pass

        elif item_type == MenuItemTypes.SEPARATOR:  # type: ignore
            layout.addSeparator()

    # Add a merge for any new/unincluded menu items (except toplevel).
    if merge:
        layout.addMerge("files")

    return layout


def model_children_to_xml(model, model_parent=None, menu_parent=None):
    """Add child menu items to menu_parent from model_parent."""
    # Prevent menu duplication that crashes gnome-menus
    global processed_directories
    processed_directories = []

    # Menus First...
    model_to_xml_menus(model, model_parent, menu_parent)

    # Includes Second... to allow for alacarte-created entries without
    # categories to persist (see LP: #1315880)
    model_to_xml_includes(model, model_parent, menu_parent)

    # Layouts Third...
    model_to_xml_layout(model, model_parent, menu_parent, merge=False)


def treeview_to_xml(treeview):
    """Write the current treeview to the -applications.menu file."""
    model = treeview.get_model()

    # Get the necessary details
    menu_name = MenuEditor.menu_name
    menu_file = MenuEditor.get_default_menu()
    merge_file = util.getSystemMenuPath(menu_file)

    if menu_file is None:
        return False

    filename = os.path.join(util.getUserMenusDirectory(), menu_file)

    # Create the menu XML
    menu = XmlMenuElementTree(menu_name, merge_file)
    root = menu.getroot()
    model_children_to_xml(model, menu_parent=root)

    # Write the file.
    menu.write(filename)

    return True